"""Convenience functions for estimating antenna radiation by applying array factor."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd
from pydantic.v1 import NonNegativeFloat, PositiveInt, conint
from scipy.signal.windows import blackman, blackmanharris, chebwin, hamming, hann, kaiser, taylor
from scipy.special import j0, jn_zeros

from tidy3d.components.base import Tidy3dBaseModel, skip_if_fields_missing
from tidy3d.components.data.monitor_data import AbstractFieldProjectionData, DirectivityData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.geometry.base import Box, Geometry
from tidy3d.components.grid.grid_spec import GridSpec, LayerRefinementSpec
from tidy3d.components.lumped_element import LumpedElement
from tidy3d.components.medium import Medium, MediumType3D
from tidy3d.components.monitor import AbstractFieldProjectionMonitor, MonitorType
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.utils import SourceType
from tidy3d.components.structure import MeshOverrideStructure, Structure
from tidy3d.components.types import TYPE_TAG_STR, ArrayLike, Axis, Bound, Undefined
from tidy3d.constants import C_0, inf
from tidy3d.exceptions import Tidy3dNotImplementedError
from tidy3d.log import log


class AbstractAntennaArrayCalculator(Tidy3dBaseModel, ABC):
    """Abstract base for phased array calculators."""

    taper: Union[RectangularTaper, RadialTaper] = pd.Field(
        None,
        discriminator=TYPE_TAG_STR,
        title="Antenna Array Taper",
        description="Amplitude weighting of array elements to control main lobe width and suppress side lobes.",
    )

    @property
    @abstractmethod
    def _antenna_locations(self) -> ArrayLike:
        """Locations of antennas' centers in an array."""

    @property
    @abstractmethod
    def _antenna_amps(self) -> ArrayLike:
        """Amplitude multipliers of antennas in an array."""

    @property
    @abstractmethod
    def _antenna_phases(self) -> ArrayLike:
        """Phase shifts of antennas in an array."""

    @property
    @abstractmethod
    def _extend_dims(self):
        """Dimensions along which antennas will be duplicated."""

    @property
    def _antenna_nominal_size(self) -> ArrayLike:
        """Size of the antenna array without taking into account size of a single antenna."""
        rmin = np.min(self._antenna_locations, axis=0)
        rmax = np.max(self._antenna_locations, axis=0)

        return rmax - rmin

    @property
    def _antenna_nominal_center(self) -> ArrayLike:
        """Center of the antenna array without taking into account size of a single antenna."""
        rmin = np.min(self._antenna_locations, axis=0)
        rmax = np.max(self._antenna_locations, axis=0)

        return 0.5 * (rmax + rmin)

    def _detect_antenna_bounds(self, simulation: Simulation):
        """Detect the bounds of the antenna in the simulation."""
        # directions in which we will need to tile simulation
        extend_dims = self._extend_dims

        # detect bounding box of all structures, sources, and lumped elements in the simulation
        sim_bounds = np.array(simulation.bounds)
        antenna_bounds = sim_bounds.copy()
        for dim in extend_dims:
            antenna_bounds[0][dim] = inf
            antenna_bounds[1][dim] = -inf

        all_objects = list(simulation.structures)
        all_objects += list(simulation.sources)
        all_objects += list(simulation.lumped_elements)

        for obj in all_objects:
            # get bounding box of the object
            if isinstance(obj, Structure):
                obj_bounds = np.array(obj.geometry.bounds)
            else:
                obj_bounds = np.array(obj.bounds)

            for dim in extend_dims:
                # update minimum and maximum bounds in each dimension
                # check if object extends beyond the simulation bounds on both sides
                extends_beyond_min = obj_bounds[0][dim] < sim_bounds[0][dim]
                extends_beyond_max = obj_bounds[1][dim] > sim_bounds[1][dim]
                if extends_beyond_min or extends_beyond_max:
                    # in case of a box, we just ignore it, since we will be able to extend it later
                    if (
                        isinstance(obj, Structure)
                        and isinstance(obj.geometry, Box)
                        and extends_beyond_min
                        and extends_beyond_max
                    ):
                        continue
                    # otherwise shrink the object bounds to simulation bounds
                    obj_bounds[0][dim] = max(obj_bounds[0][dim], sim_bounds[0][dim])
                    obj_bounds[1][dim] = min(obj_bounds[1][dim], sim_bounds[1][dim])
                    # and show a warning
                    log.warning(
                        f"Object {obj.name} (type: {obj.type}) extends beyond simulation bounds along"
                        f" '{'xyz'[dim]}' axis. Please check your antenna setup for correctness."
                    )
                # update minimum and maximum bounds in each dimension
                antenna_bounds[0][dim] = min(antenna_bounds[0][dim], obj_bounds[0][dim])
                antenna_bounds[1][dim] = max(antenna_bounds[1][dim], obj_bounds[1][dim])

        return antenna_bounds

    def _try_to_expand_geometry(
        self, geometry: Geometry, old_sim_bounds: Bound, new_sim_bounds: Bound
    ):
        """Try to expand geometry to cover the entire simulation domain."""

        can_expand = isinstance(geometry, Box) and all(
            geometry.bounds[0][dim] < old_sim_bounds[0][dim]
            or geometry.bounds[1][dim] > old_sim_bounds[1][dim]
            for dim in self._extend_dims
        )

        if not can_expand:
            return None  # could not expand geometry, so we will duplicate

        # get original structure bounds
        box_bounds = np.array(geometry.bounds)
        box_size = np.array(geometry.size)
        box_center = np.array(geometry.center)

        for dim in self._extend_dims:
            # if it extends beyond simulation bounds, but size is not inf
            # we will adjust it so that it goes out of bounds by the same amount as in the initial simulation.
            # No need to do anything if size is inf.
            if box_size[dim] != np.inf:
                # shift min bounds to new location
                box_bounds[0][dim] = new_sim_bounds[0][dim] + (
                    box_bounds[0][dim] - old_sim_bounds[0][dim]
                )

                # shift max bounds to new location
                box_bounds[1][dim] = new_sim_bounds[1][dim] - (
                    old_sim_bounds[1][dim] - box_bounds[1][dim]
                )

                # calculate new structure size and center
                box_size[dim] = box_bounds[1][dim] - box_bounds[0][dim]
                box_center[dim] = 0.5 * (box_bounds[0][dim] + box_bounds[1][dim])

        return geometry.updated_copy(
            center=tuple(box_center),
            size=tuple(box_size),
        )

    def _duplicate_or_expand_list_of_objects(
        self,
        objects: tuple[
            Union[Structure, MeshOverrideStructure, LayerRefinementSpec, LumpedElement], ...
        ],
        old_sim_bounds: Bound,
        new_sim_bounds: Bound,
    ):
        """Duplicate or expand a list of objects."""

        locations = self._antenna_locations

        array_objects = []
        for obj in objects:
            if isinstance(obj, (Structure, MeshOverrideStructure)):
                geometry = obj.geometry
                if isinstance(obj, Structure) and obj.medium.is_custom:
                    log.warning(
                        f"Object '{obj.name}' contains a custom medium which has limited support "
                        "in automatic antenna array generation. The custom medium's spatial distribution "
                        "will remain fixed and will not be translated along with the structure geometry."
                    )
            elif isinstance(obj, (LayerRefinementSpec, LumpedElement)):
                geometry = obj
            else:
                raise ValueError(f"Object of type {type(obj)} is not supported.")

            # check if it is a box and extends beyond simulation bounds
            expanded_geometry = self._try_to_expand_geometry(
                geometry=geometry, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
            )

            if expanded_geometry is None:
                # could not expand geometry, so we duplicate
                for ind, translation_vector in enumerate(locations):
                    if isinstance(obj, Structure):
                        new_obj = obj.updated_copy(
                            geometry=obj.geometry.translated(*translation_vector),
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    elif isinstance(obj, MeshOverrideStructure):
                        new_obj = obj.updated_copy(
                            geometry=obj.geometry.translated(*translation_vector).bounding_box,
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    elif isinstance(obj, LayerRefinementSpec):
                        new_obj = obj.updated_copy(
                            center=tuple(obj.center + translation_vector),
                        )
                    elif isinstance(obj, LumpedElement):
                        new_obj = obj.updated_copy(
                            center=tuple(obj.center + translation_vector),
                            name=None if obj.name is None else f"{obj.name}_{ind}",
                        )
                    array_objects.append(new_obj)
            else:
                # could expand geometry, so we create a copy of the original object with updated size and center
                if isinstance(obj, (Structure, MeshOverrideStructure)):
                    new_obj = obj.updated_copy(
                        geometry=expanded_geometry,
                    )
                elif isinstance(obj, (LayerRefinementSpec, LumpedElement)):
                    new_obj = expanded_geometry

                # add the new object to the list of objects
                array_objects.append(new_obj)

        return array_objects

    def _expand_monitors(
        self,
        monitors: tuple[MonitorType, ...],
        antenna_bounds: Bound,
        new_sim_bounds: Bound,
        old_sim_bounds: Bound,
    ):
        """Expand monitors."""

        extend_dims = self._extend_dims

        # expand far-field monitors
        array_monitors = []
        for monitor in monitors:
            # we only expand field projection monitors
            if isinstance(monitor, AbstractFieldProjectionMonitor):
                # get original monitor bounds
                mnt_bounds = np.array(monitor.bounds)
                mnt_size = np.array(monitor.size)
                mnt_center = np.array(monitor.center)

                if any(mnt_size[dim] == 0 for dim in extend_dims):
                    log.warning(
                        f"Monitor '{monitor.name}' (type: '{monitor.type}') has zero size along "
                        f"one of the axes. It will not be included in the resulting simulation."
                    )
                    continue

                for dim in extend_dims:
                    if mnt_size[dim] != np.inf:
                        # check that monitor covers the estimated antenna box
                        if (
                            mnt_bounds[0][dim] > antenna_bounds[0][dim]
                            or mnt_bounds[1][dim] < antenna_bounds[1][dim]
                        ):
                            log.warning(
                                f"Monitor '{monitor.name}' (type: '{monitor.type}') does not cover "
                                f"the estimated antenna box along '{'xyz'[dim]}' axis. "
                                "The automatically extended monitor will likely be wrong. "
                                "Please double check the resulting simulation."
                            )
                        # shift min bounds to new location
                        mnt_bounds[0][dim] = new_sim_bounds[0][dim] + (
                            mnt_bounds[0][dim] - old_sim_bounds[0][dim]
                        )

                        # shift max bounds to new location
                        mnt_bounds[1][dim] = new_sim_bounds[1][dim] - (
                            old_sim_bounds[1][dim] - mnt_bounds[1][dim]
                        )

                        # calculate new monitor size and center
                        mnt_size[dim] = mnt_bounds[1][dim] - mnt_bounds[0][dim]
                        mnt_center[dim] = 0.5 * (mnt_bounds[0][dim] + mnt_bounds[1][dim])

                # create a copy of the original monitor with updated size and center
                new_mnt = monitor.updated_copy(
                    center=tuple(mnt_center),
                    size=tuple(mnt_size),
                )

                # add the new monitor to the list of monitors
                array_monitors.append(new_mnt)

            # otherwise we ignore and warn the user
            else:
                log.warning(
                    f"Monitor '{monitor.name}' (type: '{monitor.type}') will not be automatically "
                    "transferred into the resulting antenna array simulation."
                )

        return array_monitors

    def _duplicate_structures(
        self, structures: tuple[Structure, ...], new_sim_bounds: Bound, old_sim_bounds: Bound
    ):
        """Duplicate structures."""

        return self._duplicate_or_expand_list_of_objects(
            objects=structures, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
        )

    def _duplicate_sources(
        self,
        sources: tuple[SourceType, ...],
        lumped_elements: tuple[LumpedElement, ...],
        old_sim_bounds: Bound,
        new_sim_bounds: Bound,
    ):
        """Duplicate sources and lumped elements."""
        array_lumped_elements = self._duplicate_or_expand_list_of_objects(
            objects=lumped_elements, old_sim_bounds=old_sim_bounds, new_sim_bounds=new_sim_bounds
        )

        array_sources = []
        for ind, (translation_vector, phase_shift, amp_multiplier) in enumerate(
            zip(self._antenna_locations, self._antenna_phases, self._antenna_amps)
        ):
            if amp_multiplier != 0.0:
                for source in sources:
                    new_source = source.updated_copy(
                        center=tuple(source.center + translation_vector),
                        name=f"{source.name}_{ind}",
                        source_time=source.source_time.updated_copy(
                            phase=source.source_time.phase + phase_shift,
                            amplitude=source.source_time.amplitude * amp_multiplier,
                        ),
                    )
                    array_sources.append(new_source)

        return array_sources, array_lumped_elements

    def _duplicate_grid_specs(
        self, grid_spec: GridSpec, old_sim_bounds: Bound, new_sim_bounds: Bound
    ):
        """Duplicate grid specs."""

        array_overrides = self._duplicate_or_expand_list_of_objects(
            objects=grid_spec.override_structures,
            old_sim_bounds=old_sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )
        array_layer_refinement_specs = self._duplicate_or_expand_list_of_objects(
            objects=grid_spec.layer_refinement_specs,
            old_sim_bounds=old_sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )

        array_snapping_points = []
        for translation_vector in self._antenna_locations:
            for snapping_point in grid_spec.snapping_points:
                new_snapping_point = [
                    snapping_point[dim] + translation_vector[dim]
                    if snapping_point[dim] is not None
                    else None
                    for dim in range(3)
                ]
                array_snapping_points.append(new_snapping_point)

        return grid_spec.updated_copy(
            override_structures=array_overrides,
            layer_refinement_specs=array_layer_refinement_specs,
            snapping_points=array_snapping_points,
        )

    def make_antenna_array(self, simulation: Simulation) -> Simulation:
        """
        Converts a single antenna simulation into an antenna array simulation.
        This function identifies the size and position of the single antenna
        in the input simulation and uses this information to compute the dimensions
        of the resulting antenna array simulation. All structures, sources, lumped elements,
        and mesh override structures are duplicated, while monitors are extended in size.
        Only projection monitors are transferred into the resulting simulation.

        For best results, the antenna assembly should be contained within the simulation bounds.

        Parameters:
        ----------
        simulation : Simulation
            The simulation specification describing a single antenna setup.

        Returns:
        --------
        Simulation
            The simulation specification describing the antenna array.
        """

        # detect bounding box of all structures, sources, and lumped elements in the simulation
        sim_bounds = np.array(simulation.bounds)
        antenna_bounds = self._detect_antenna_bounds(simulation)

        # compute the center and size of the antenna
        antenna_size = antenna_bounds[1] - antenna_bounds[0]
        antenna_center = 0.5 * (antenna_bounds[0] + antenna_bounds[1])

        # compute buffer between the antenna and simulation bounds on each side
        buffer_min = antenna_bounds[0] - sim_bounds[0]
        buffer_max = sim_bounds[1] - antenna_bounds[1]

        # compute total size of the antenna array in x, y, and z directions
        antenna_array_size = self._antenna_nominal_size + antenna_size

        new_sim_bounds = [
            antenna_center + self._antenna_nominal_center - antenna_array_size / 2 - buffer_min,
            antenna_center + self._antenna_nominal_center + antenna_array_size / 2 + buffer_max,
        ]

        # compute the total size of the simulation domain
        new_sim_size = new_sim_bounds[1] - new_sim_bounds[0]
        new_sim_center = 0.5 * (new_sim_bounds[0] + new_sim_bounds[1])

        # duplicate structures, sources, lumped elements, and override structures
        array_structures = self._duplicate_structures(
            structures=simulation.structures,
            new_sim_bounds=new_sim_bounds,
            old_sim_bounds=sim_bounds,
        )
        array_sources, array_lumped_elements = self._duplicate_sources(
            sources=simulation.sources,
            lumped_elements=simulation.lumped_elements,
            old_sim_bounds=sim_bounds,
            new_sim_bounds=new_sim_bounds,
        )
        array_monitors = self._expand_monitors(
            monitors=simulation.monitors,
            antenna_bounds=antenna_bounds,
            new_sim_bounds=new_sim_bounds,
            old_sim_bounds=sim_bounds,
        )
        array_grid_spec = self._duplicate_grid_specs(
            grid_spec=simulation.grid_spec, old_sim_bounds=sim_bounds, new_sim_bounds=new_sim_bounds
        )

        new_sim = simulation.updated_copy(
            center=tuple(new_sim_center),
            size=tuple(new_sim_size),
            structures=array_structures,
            monitors=array_monitors,
            sources=array_sources,
            lumped_elements=array_lumped_elements,
            grid_spec=array_grid_spec,
        )

        return new_sim

    @abstractmethod
    def array_factor(
        self,
        theta: Union[float, ArrayLike],
        phi: Union[float, ArrayLike],
        frequency: Union[NonNegativeFloat, ArrayLike],
    ) -> ArrayLike:
        """
        Compute the array factor for an antenna array.

        Parameters:
        -----------
        theta : Union[float, ArrayLike]
            Observation angles in the elevation plane (in radians).
        phi : Union[float, ArrayLike]
            Observation angles in the azimuth plane (in radians).
        frequency : Union[NonNegativeFloat, ArrayLike]
            Signal frequency (in Hz).

        Returns:
        --------
        ArrayLike
            Array factor values for each combination of theta and phi.
        """

    def monitor_data_from_array_factor(
        self,
        monitor_data: AbstractFieldProjectionData,
        new_monitor: AbstractFieldProjectionMonitor = None,
    ) -> AbstractFieldProjectionData:
        """Apply the array factor to the monitor data of a single antenna.

        Parameters
        ----------
        monitor_data : AbstractFieldProjectionData
            The monitor data of a single antenna.
        new_monitor : AbstractFieldProjectionMonitor = None
            The new monitor to be used in the resulting data.

        Returns
        -------
        AbstractFieldProjectionData
            The monitor data of the antenna array.
        """

        # Get spherical coordinates
        r, theta, phi = list(monitor_data.coords_spherical.values())
        freqs = monitor_data.f
        coords_shape = list(np.shape(r))
        coords_shape.append(len(freqs))

        # Compute the array factor
        af = self.array_factor(theta.ravel(), phi.ravel(), freqs, monitor_data.medium)
        af = np.reshape(af, coords_shape)

        update_dict = {}

        # Apply the array factor to the monitor data
        for key, field in monitor_data.field_components.items():
            update_dict[key] = field.copy() * af

        if new_monitor is not None:
            monitor = new_monitor
        else:
            monitor = monitor_data.monitor

        update_dict["monitor"] = monitor
        update_dict["projection_surfaces"] = monitor.projection_surfaces

        if isinstance(monitor_data, DirectivityData):
            # Attempt to recompute flux
            try:
                update_dict["flux"] = monitor_data.flux_from_projected_fields()
            except ValueError as e:
                log.warning(
                    "Could not recalculate flux for a 'DirectivityData' due to the following "
                    f"reason: {e} This monitor will not be included in the resulting data."
                )
                return None

        # Create a new monitor data with the updated fields
        new_mnt_data = monitor_data.updated_copy(
            **update_dict,
        )

        return new_mnt_data

    def simulation_data_from_array_factor(
        self,
        antenna_data: SimulationData,
    ) -> SimulationData:
        """
        Computes the far-field data of a rectangular antenna array based on the far-field data of
        a single antenna. Additionaly, it automatically converts a single antenna simulation setup
        into the corresponding antenna array simulation setup.

        Note that any near-field monitor data will be ignored.

        Parameters
        ----------
        antenna_data : SimulationData
            The far-field data of a single antenna.

        Returns
        -------
        SimulationData
            The far-field data of the antenna array.
        """

        # create an expanded simulation for reference
        sim_array = self.make_antenna_array(antenna_data.simulation)

        # names of transferred monitors
        mnt_dict = {mnt.name: mnt for mnt in sim_array.monitors}

        # process far field data
        data_array = []
        good_monitors = []
        for mnt_data in antenna_data.data:
            mnt_name = mnt_data.monitor.name
            if mnt_name in mnt_dict:
                array_mnt_data = self.monitor_data_from_array_factor(
                    monitor_data=mnt_data,
                    new_monitor=mnt_dict[mnt_name],
                )
                if array_mnt_data is not None:
                    good_monitors.append(mnt_dict[mnt_name])
                    data_array.append(array_mnt_data)

        return SimulationData(
            simulation=sim_array.updated_copy(monitors=good_monitors), data=data_array
        )

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values

    def _rect_taper_array_factor(
        self, exp_x: ArrayLike, exp_y: ArrayLike, exp_z: ArrayLike
    ) -> ArrayLike:
        """
        Compute the array factor assuming a separable rectangular (Cartesian) taper.

        This method evaluates the array factor using separable amplitude weights
        along the x, y, and z dimensions.

        Parameters
        ----------
        exp_x: ArrayLike
            3D array of phases along x axis [N_x, N_theta, N_freq].
        exp_y: ArrayLike
            3D array of phases along x axis [N_y, N_theta, N_freq].
        exp_z: ArrayLike
            3D array of phases along x axis [N_z, N_theta, N_freq].

        Returns
        -------
        ArrayLike
            Array factor values for each combination of theta and phi.
        """

        # if taper is not defined set all amplitudes to 1.0
        if self.taper is None:
            amp_x = (
                1.0 if self.amp_multipliers[0] is None else self.amp_multipliers[0][:, None, None]
            )
            amp_y = (
                1.0 if self.amp_multipliers[1] is None else self.amp_multipliers[1][:, None, None]
            )
            amp_z = (
                1.0 if self.amp_multipliers[2] is None else self.amp_multipliers[2][:, None, None]
            )

        # if rectangular taper is spacified
        elif isinstance(self.taper, RectangularTaper):
            # get amplitudes along x, y and z axes
            amp_x, amp_y, amp_z = self.taper.amp_multipliers(self.array_size)

            # broadcast amplitudes to [N_x,1,1], [N_y,1,1] and [Nz,1,1], respectively
            amp_x = amp_x[:, None, None]
            amp_y = amp_y[:, None, None]
            amp_z = amp_z[:, None, None]

        # Tapers with non-separable amplitude weights are not supported by this function
        else:
            raise ValueError(f"Unsupported taper type {type(self.taper)} was passed.")

        # Calculate individual array factors in x, y, and z directions
        af_x = np.sum(
            amp_x * exp_x,
            axis=0,
        )
        af_y = np.sum(
            amp_y * exp_y,
            axis=0,
        )
        af_z = np.sum(
            amp_z * exp_z,
            axis=0,
        )

        # Calculate the overall array factor
        array_factor = af_x * af_y * af_z

        return array_factor

    def _general_taper_array_factor(
        self, exp_x: ArrayLike, exp_y: ArrayLike, exp_z: ArrayLike
    ) -> ArrayLike:
        """
        Compute the array factor assuming a non-separable (non-Cartesian) taper.

        This method evaluates the array factor using non-separable amplitude weights
        along the x, y, and z dimensions.

        Parameters
        ----------
        exp_x: ArrayLike
            3D array of phases along x axis [N_x, N_theta, N_freq].
        exp_y: ArrayLike
            3D array of phases along x axis [N_y, N_theta, N_freq].
        exp_z: ArrayLike
            3D array of phases along x axis [N_z, N_theta, N_freq].

        Returns
        -------
        ArrayLike
            Array factor values for each combination of theta / phi and frequency.
        """
        # get taper weights
        amps = self.taper.amp_multipliers(self.array_size)

        # ensure amplitude weights are in format tuple[ArrayLike, ]
        if len(amps) != 1:
            raise ValueError(
                "Non-cartesian taper was expected. Please ensure a valid taper is used."
            )

        # compute array factor: AF(theta,f) = sum_{x,y,z} amp(x,y,z) * exp_x(x,theta,f)*exp_y(y,theta,f)*exp_z(z,theta,f)
        array_factor = np.einsum("xpf,ypf,zpf,xyz->pf", exp_x, exp_y, exp_z, amps[0])

        return array_factor


class RectangularAntennaArrayCalculator(AbstractAntennaArrayCalculator):
    """This class provides methods to calculate the array factor and far-field radiation patterns
    for rectangular phased antenna arrays. It handles arrays with arbitrary size, spacing,
    phase shifts, and amplitude tapering in x, y, and z directions.

    Notes
    -----

    The array factor is calculated using the standard array factor formula for rectangular arrays,
    which accounts for the spatial distribution of antennas and their relative phases and amplitudes.
    This can be used to analyze beam steering, sidelobe levels, and other array characteristics.

    In addition, this class provides a convenience method to create an antenna array simulation
    from a single antenna simulation. This can be used to compute the behavior (near-field and/or
    far-field) of the full antenna array directly without any approximations. Such a simulation setup
    can be obtained by directly calling the `make_antenna_array` function, or by accessing the field `.simulation`
    of the :class:`SimulationData` object returned by the `simulation_data_from_array_factor` method.

    Example:
    --------
    >>> array_calculator = RectangularAntennaArrayCalculator(
    ...    array_size=(3, 4, 5),
    ...    spacings=(0.5, 0.5, 0.5),
    ...    phase_shifts=(0, 0, 0),
    ... ) # doctest: +SKIP
    """

    array_size: tuple[PositiveInt, PositiveInt, PositiveInt] = pd.Field(
        title="Array Size",
        description="Number of antennas along x, y, and z directions.",
    )

    spacings: tuple[NonNegativeFloat, NonNegativeFloat, NonNegativeFloat] = pd.Field(
        title="Antenna Spacings",
        description="Center-to-center spacings between antennas along x, y, and z directions.",
    )

    phase_shifts: tuple[float, float, float] = pd.Field(
        (0, 0, 0),
        title="Phase Shifts",
        description="Phase-shifts between antennas along x, y, and z directions.",
    )

    amp_multipliers: tuple[Optional[ArrayLike], Optional[ArrayLike], Optional[ArrayLike]] = (
        pd.Field(
            (None, None, None),
            title="Amplitude Multipliers",
            description="Amplitude multipliers spatially distributed along x, y, and z directions.",
        )
    )

    @pd.validator("amp_multipliers", pre=True, always=True)
    @skip_if_fields_missing(["array_size"])
    def _check_amp_multipliers(cls, val, values):
        """Check that the length of the amplitude multipliers is equal to the array size along each dimension."""
        array_size = values.get("array_size")
        if len(val) != 3:
            raise ValueError("'amp_multipliers' must have 3 elements.")
        if val[0] is not None and len(val[0]) != array_size[0]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[0])} along the x direction, but the array size is {array_size[0]}."
            )
        if val[1] is not None and len(val[1]) != array_size[1]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[1])} along the y direction, but the array size is {array_size[1]}."
            )
        if val[2] is not None and len(val[2]) != array_size[2]:
            raise ValueError(
                f"'amp_multipliers' has length of {len(val[2])} along the z direction, but the array size is {array_size[2]}."
            )
        return val

    @property
    def _antenna_locations(self) -> ArrayLike:
        """Locations of antennas' centers in an array."""

        x = (np.arange(self.array_size[0]) - (self.array_size[0] - 1) / 2) * self.spacings[0]
        y = (np.arange(self.array_size[1]) - (self.array_size[1] - 1) / 2) * self.spacings[1]
        z = (np.arange(self.array_size[2]) - (self.array_size[2] - 1) / 2) * self.spacings[2]

        X, Y, Z = np.meshgrid(x, y, z, indexing="ij")

        return np.transpose([X.ravel(), Y.ravel(), Z.ravel()])

    @property
    def _antenna_amps(self) -> ArrayLike:
        """Amplitude multipliers of antennas in an array."""

        if self.taper is not None:
            if isinstance(self.taper, RectangularTaper):
                amp_x, amp_y, amp_z = self.taper.amp_multipliers(self.array_size)

                # broadcast amplitudes to [N_x,1,1], [N_y,1,1] and [Nz,1,1], respectively
                amp_x = amp_x[:, None, None]
                amp_y = amp_y[None, :, None]
                amp_z = amp_z[None, None, :]

                amps = amp_x * amp_y * amp_z

            else:
                amps = self.taper.amp_multipliers(self.array_size)

            return np.ravel(amps)

        amps_per_dim = [
            np.ones(size) if multiplier is None else multiplier
            for multiplier, size in zip(self.amp_multipliers, self.array_size)
        ]

        amps_grid = np.meshgrid(*amps_per_dim, indexing="ij")

        return np.ravel(amps_grid[0] * amps_grid[1] * amps_grid[2])

    @property
    def _antenna_phases(self) -> ArrayLike:
        """Phase shifts of antennas in an array."""

        phase_shifts_per_dim = [
            np.arange(self.array_size[dim]) * self.phase_shifts[dim] for dim in range(3)
        ]

        phase_shifts_grid = np.meshgrid(*phase_shifts_per_dim, indexing="ij")

        return np.ravel(sum(p for p in phase_shifts_grid))

    @property
    def _extend_dims(self) -> tuple[Axis, ...]:
        """Dimensions along which antennas will be duplicated."""
        return [ind for ind, size in enumerate(self.array_size) if size > 1]

    def array_factor(
        self,
        theta: Union[float, ArrayLike],
        phi: Union[float, ArrayLike],
        frequency: Union[NonNegativeFloat, ArrayLike],
        medium: MediumType3D = Undefined,
    ) -> ArrayLike:
        """
        Compute the array factor for a 3D antenna array.

        Parameters
        ----------
        theta : Union[float, ArrayLike]
            Observation angles in the elevation plane (in radians).
        phi : Union[float, ArrayLike]
            Observation angles in the azimuth plane (in radians).
        frequency : Union[NonNegativeFloat, ArrayLike]
            Signal frequency (in Hz).

        Returns
        -------
        ArrayLike
            Array factor values for each combination of azimuth and zenith angles.
        """
        if medium is Undefined:
            medium = Medium()

        # Convert all inputs to numpy arrays
        theta_array = np.atleast_1d(theta)
        phi_array = np.atleast_1d(phi)

        # ensure that theta and phi have the same length
        if len(theta_array) != len(phi_array):
            raise ValueError("'theta' and 'phi' must have the same length.")

        # reshape inputs for easier broadcasting
        theta_array = np.reshape(theta_array, (len(theta_array), 1))
        phi_array = np.reshape(phi_array, (len(phi_array), 1))
        f_array = np.reshape(frequency, (1, len(np.atleast_1d(frequency))))

        eps_complex = np.array(medium.eps_model(f_array))

        wavelength = C_0 / f_array / np.sqrt(eps_complex)
        k = 2 * np.pi / wavelength  # Wavenumber

        # Calculate the phase shift in the x, y, and z directions
        psi_x = (
            k * self.spacings[0] * np.sin(theta_array) * np.cos(phi_array) - self.phase_shifts[0]
        )
        psi_y = (
            k * self.spacings[1] * np.sin(theta_array) * np.sin(phi_array) - self.phase_shifts[1]
        )
        psi_z = k * self.spacings[2] * np.cos(theta_array) - self.phase_shifts[2]

        # Calculate resulting complex exponentials
        exp_x = np.exp(-1j * np.arange(self.array_size[0])[:, None, None] * psi_x[np.newaxis, :])
        exp_y = np.exp(-1j * np.arange(self.array_size[1])[:, None, None] * psi_y[np.newaxis, :])
        exp_z = np.exp(-1j * np.arange(self.array_size[2])[:, None, None] * psi_z[np.newaxis, :])

        # Compute array factor based on the defined taper
        if self.taper is None or isinstance(self.taper, RectangularTaper):
            return self._rect_taper_array_factor(exp_x, exp_y, exp_z)
        else:
            return self._general_taper_array_factor(exp_x, exp_y, exp_z)


class AbstractWindow(Tidy3dBaseModel, ABC):
    """This class provides interface for window selection."""

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """Interface function for computing window weights at N points."""
        raise Tidy3dNotImplementedError(
            f"Calculation of antenna amplitudes at a discrete number of points is not yet implemented for window type {self.type}."
        )

    def _get_weights_continuous(self, p_vec: ArrayLike) -> ArrayLike:
        """Interface function for computing window weights at given locations."""
        raise Tidy3dNotImplementedError(
            f"Calculation of antenna amplitudes at arbitrary locations is not yet implemented for window type {self.type}."
        )


class HammingWindow(AbstractWindow):
    """Standard Hamming window for tapering or spectral shaping."""

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Hamming window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Hamming window weights.
        """
        return hamming(N)


class BlackmanWindow(AbstractWindow):
    """Standard Blackman window for tapering or spectral shaping."""

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Blackman window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Blackman window weights.
        """
        return blackman(N)


class BlackmanHarrisWindow(AbstractWindow):
    """Standard Blackman-Harris window for tapering or spectral shaping."""

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Blackman-Harris window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Blackman-Harris window weights.
        """
        return blackmanharris(N)


class HannWindow(AbstractWindow):
    """Hann window with configurable sidelobe suppression and sidelobe count."""

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Hann window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Hann window weights.
        """
        return hann(N)


class ChebWindow(AbstractWindow):
    """Standard Chebyshev window for tapering with configurable sidelobe attenuation."""

    attenuation: pd.PositiveFloat = pd.Field(
        default=30,
        title="Attenuation",
        description="Desired attenuation level of sidelobes.",
        units="dB",
    )

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Chebyshev window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Chebyshev window weights.
        """
        return chebwin(N, self.attenuation)


class KaiserWindow(AbstractWindow):
    """Class for Kaiser window."""

    beta: pd.NonNegativeFloat = pd.Field(
        ...,
        title="Shape Parameter",
        description="Shape parameter, determines trade-off between main-lobe width and side lobe level.",
    )

    def _get_weights_discrete(self, N: int) -> ArrayLike:
        """
        Generate a 1D Kaiser window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Kaiser window weights.
        """
        return kaiser(N, self.beta)


class TaylorWindow(AbstractWindow):
    """Taylor window with configurable sidelobe suppression and sidelobe count."""

    sll: pd.PositiveFloat = pd.Field(
        default=30,
        title="Sidelobe Suppression Level",
        description="Desired suppression of sidelobe level relative to the DC gain.",
        units="dB",
    )

    nbar: conint(gt=0, le=10) = pd.Field(
        default=4,
        title="Number of Nearly Constant Sidelobes",
        description="Number of nearly constant level sidelobes adjacent to the mainlobe.",
    )

    def _get_weights_discrete(self, N):
        """
        Generate a 1D Taylor window of length N.

        Parameters
        ----------
        N : int
            Number of points in the window.

        Returns
        -------
        ArrayLike
            1D array of Taylor window weights.
        """
        return taylor(N, self.nbar, self.sll)

    def _get_exp_weights(self, mus: np.ndarray):
        """
        Compute expansion coefficients B_l for the circular Taylor taper.

        The aperture field E_a(p) is represented as a sum over Bessel functions:
            E_a(p) ≈ 1 + sum_l B_l * J0(mu_l * p)

        where mu_l are the zeros of J1(pi * mu), and B_l are the expansion coefficients
        chosen to enforce a specified sidelobe level (sll) via the Taylor design method.

        Parameters:
        -----------
        mus : np.ndarray
            Roots of J1(pi * mu) / pi (used to construct the Bessel function basis).

        Returns:
        --------
        B : np.ndarray
            Expansion coefficients (1D array of shape (nbar - 1,)).
        """

        # calculate real-valued parameter from sidelobe attenuation level
        A = np.arccosh(10 ** (self.sll / 20)) / np.pi
        sigma = mus[-1] / np.sqrt(A**2 + (self.nbar - 0.5) ** 2)

        u = np.sqrt(A**2 + (np.arange(1, self.nbar) - 0.5) ** 2)
        B = np.zeros(self.nbar - 1)

        for i in range(self.nbar - 1):
            mu_i_sq = mus[i] ** 2

            # Numerator: product over (1 - mu_i^2 / (sigma * u_n)^2)
            num_terms = 1 - mu_i_sq / (sigma * u) ** 2
            num = np.prod(num_terms)

            # Denominator
            denom_terms = [1 - mu_i_sq / mus[n] ** 2 for n in range(self.nbar - 1) if n != i]
            denom = np.prod(denom_terms)
            B[i] = -num / (denom * j0(np.pi * mus[i]))

        return B

    def _get_weights_continuous(self, p_vec: ArrayLike) -> ArrayLike:
        """
        Sample weights from the circular Taylor taper at specified radial positions.

        Parameters
        ----------
        p_vec : ArrayLike
            1D array of radial sampling points in the range [0, π].

        Returns
        -------
        g_p_norm : ArrayLike
            1D array of Taylor taper weights evaluated at the points in ``p_vec``.
        """

        # get locations J1(np.pi * mu)=0
        mus = jn_zeros(1, self.nbar) / np.pi

        B_m = self._get_exp_weights(mus=mus)

        J = j0(np.outer(mus[0:-1], p_vec))

        g_p = 1 + B_m @ J

        return g_p


# define a list of acceptable rectangular windows
RectangularWindowType = Union[
    HammingWindow,
    HannWindow,
    KaiserWindow,
    TaylorWindow,
    ChebWindow,
    BlackmanWindow,
    BlackmanHarrisWindow,
]


class AbstractTaper(Tidy3dBaseModel, ABC):
    """Abstract taper class provides an interface for taper of Array antennas."""

    @abstractmethod
    def amp_multipliers(
        self, array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
    ) -> tuple[np.ndarray, ...]:
        """
        Compute taper amplitudes for phased array antennas.

        Parameters:
        ----------
        array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
            A tuple of array size along x,y, and z axes.
        """


class RectangularTaper(AbstractTaper):
    """Class for rectangular taper."""

    window_x: Optional[RectangularWindowType] = pd.Field(
        None,
        title="X Axis Window",
        description="Window type used to taper array antenna along x axis.",
        discriminator=TYPE_TAG_STR,
    )

    window_y: Optional[RectangularWindowType] = pd.Field(
        None,
        title="Y Axis Window",
        description="Window type used to taper array antenna along y axis.",
        discriminator=TYPE_TAG_STR,
    )

    window_z: Optional[RectangularWindowType] = pd.Field(
        None,
        title="Z Axis Window",
        description="Window type used to taper array antenna along z axis.",
        discriminator=TYPE_TAG_STR,
    )

    @classmethod
    def from_isotropic_window(cls, window: RectangularWindowType) -> RectangularTaper:
        """
        Set the same window along x, y, and z dimensions.

        Parameters
        ----------
        window: RectangularWindowType
            A supported 1D window type from ``RectangularWindowType``.

        Returns
        -------
        RectangularTaper
            A ``RectangularTaper`` instance with all three dimensions set to the specified window.
        """
        return cls(window_x=window, window_y=window, window_z=window)

    @pd.root_validator
    def check_at_least_one_window(cls, values):
        if not any([values.get("window_x"), values.get("window_y"), values.get("window_z")]):
            raise ValueError("At least one window (x, y, or z) must be provided.")
        return values

    def amp_multipliers(
        self, array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
    ) -> tuple[ArrayLike, ArrayLike, ArrayLike]:
        """
        Method ``amp_multipliers()`` computes rectangular taper amplitude for phased array antennas.

        Parameters:
        ----------
        array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
            A tuple of array size along x,y, and z axes.

        Returns:
        --------
        tuple[ArrayLike, ArrayLike, ArrayLike]
            a tuple of three 1D numpy arrays with taper amplitudes along x,y, and z axes.
        """

        effective_size = tuple(dim if dim is not None else 1 for dim in array_size)

        amps = (
            window._get_weights_discrete(effective_size[ind])
            if window is not None
            else np.ones(effective_size[ind])
            for ind, window in enumerate([self.window_x, self.window_y, self.window_z])
        )

        return amps


class RadialTaper(AbstractTaper):
    """Class for Radial Taper."""

    window: TaylorWindow = pd.Field(
        ..., title="Window Object", description="Window type used to taper array antenna."
    )

    def amp_multipliers(
        self, array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
    ) -> tuple[ArrayLike,]:
        """
        Method ``amp_multipliers()`` computes radial taper amplitude for phased array antennas.

        Parameters:
        ----------
        array_size: tuple[PositiveInt, PositiveInt, PositiveInt]
            A tuple of array size along x,y, and z axes.

        Returns:
        --------
        tuple[ArrayLike,]
            a tuple of one 3D numpy array with taper amplitudes.
        """
        effective_size = tuple(dim if dim is not None else 1 for dim in array_size)

        # Generate grid of indices
        grid = np.indices(effective_size)
        idx_c = np.array(effective_size) // 2

        # Compute distances to center
        dists = np.linalg.norm(grid - idx_c[:, None, None, None], axis=0)

        norm_dists = dists / np.max(dists) * np.pi

        amps = self.window._get_weights_continuous(norm_dists)

        amps = np.reshape(amps, effective_size)

        return (amps,)


RectangularAntennaArrayCalculator.update_forward_refs()



from __future__ import annotations

import autograd.numpy as np
from numpy.typing import NDArray


def grey_indicator(array: NDArray) -> float:
    """Calculate the grey indicator for a given array.

    The grey indicator returns 1 for completely grey arrays (all 0.5) and 0 for
    perfectly binary arrays. It is calculated based on:
    Sigmund, Ole. "Morphology-based black and white filters for topology optimization."
    Structural and Multidisciplinary Optimization 33 (2007): 401-424.

    Parameters
    ----------
    array : np.ndarray
        The input array for which the grey indicator is to be calculated.

    Returns
    -------
    float
        The calculated grey indicator.
    """
    return np.mean(4 * array * (1 - array))



"""Monitor data for unstructured volume mesh monitors."""

from __future__ import annotations

from typing import Union

import pydantic.v1 as pd

from tidy3d.components.data.utils import TetrahedralGridDataset, TriangularGridDataset
from tidy3d.components.tcad.data.monitor_data.abstract import HeatChargeMonitorData
from tidy3d.components.tcad.monitors.mesh import VolumeMeshMonitor

UnstructuredFieldType = Union[TriangularGridDataset, TetrahedralGridDataset]


class VolumeMeshData(HeatChargeMonitorData):
    """Data associated with a :class:`VolumeMeshMonitor`: stores the unstructured mesh.

    Example
    -------

    >>> import tidy3d as td
    >>> import numpy as np
    >>> mesh_mnt = td.VolumeMeshMonitor(size=(1, 2, 3), name="mesh")
    >>> tet_grid_points = td.PointDataArray(
    ...     [[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]],
    ...     dims=("index", "axis"),
    ... )
    >>> tet_grid_cells = td.CellDataArray(
    ...     [[0, 1, 2, 4], [1, 2, 3, 4]],
    ...     dims=("cell_index", "vertex_index"),
    ... )
    >>> tet_grid_values = td.IndexedDataArray(
    ...     np.zeros((tet_grid_points.shape[0],)),
    ...     dims=("index",),
    ...     name="Mesh",
    ... )
    >>> tet_grid = td.TetrahedralGridDataset(
    ...     points=tet_grid_points,
    ...     cells=tet_grid_cells,
    ...     values=tet_grid_values,
    ... )
    >>> mesh_mnt_data = td.VolumeMeshData(monitor=mesh_mnt, mesh=tet_grid) # doctest: +SKIP
    """

    monitor: VolumeMeshMonitor = pd.Field(
        ..., title="Monitor", description="Mesh monitor associated with the data."
    )

    mesh: UnstructuredFieldType = pd.Field(
        ...,
        title="Mesh",
        description="Dataset storing the mesh.",
    )

    @property
    def field_components(self) -> dict[str, UnstructuredFieldType]:
        """Maps the field components to their associated data."""
        return {"mesh": self.mesh}

    def field_name(self, val: str) -> str:
        """Gets the name of the fields to be plot."""
        return "Mesh"



from __future__ import annotations

from typing import Optional

import numpy as np

from tidy3d.components.microwave.data.monitor_data import AntennaMetricsData
from tidy3d.plugins.smatrix.analysis.terminal import (
    compute_wave_amplitudes_at_each_port,
)
from tidy3d.plugins.smatrix.data.data_array import PortDataArray
from tidy3d.plugins.smatrix.data.terminal import TerminalComponentModelerData


def get_antenna_metrics_data(
    terminal_component_modeler_data: TerminalComponentModelerData,
    port_amplitudes: Optional[dict[str, complex]] = None,
    monitor_name: Optional[str] = None,
) -> AntennaMetricsData:
    """Calculate antenna parameters using superposition of fields from multiple port excitations.

    The method computes the radiated far fields and port excitation power wave amplitudes
    for a superposition of port excitations, which can be used to analyze antenna radiation
    characteristics.

    Parameters
    ----------
    terminal_component_modeler_data: TerminalComponentModelerData
        Data associated with a :class:`.TerminalComponentModeler` simulation run.
    port_amplitudes : dict[str, complex] = None
        Dictionary mapping port names to their desired excitation amplitudes. For each port,
        :math:`\\frac{1}{2}|a|^2` represents the incident power from that port into the system.
        If None, uses only the first port without any scaling of the raw simulation data.
    monitor_name : str = None
        Name of the :class:`.DirectivityMonitor` to use for calculating far fields.
        If None, uses the first monitor in `radiation_monitors`.

    Returns
    -------
    :class:`.AntennaMetricsData`
        Container with antenna parameters including directivity, gain, and radiation efficiency,
        computed from the superposition of fields from all excited ports.
    """
    # Use the first port as default if none specified
    if port_amplitudes is None:
        port_amplitudes = {terminal_component_modeler_data.modeler.ports[0].name: None}
    # Check port names, and create map from port to amplitude
    port_dict = {}
    for key in port_amplitudes.keys():
        port, _ = terminal_component_modeler_data.modeler.network_dict[key]
        port_dict[port] = port_amplitudes[key]
    # Get the radiation monitor, use first as default
    # if none specified
    if monitor_name is None:
        rad_mon = terminal_component_modeler_data.modeler.radiation_monitors[0]
    else:
        rad_mon = terminal_component_modeler_data.modeler.get_radiation_monitor_by_name(
            monitor_name
        )

    # Create data arrays for holding the superposition of all port power wave amplitudes
    f = list(rad_mon.freqs)
    coords = {"f": f, "port": list(terminal_component_modeler_data.modeler.matrix_indices_monitor)}
    a_sum = PortDataArray(
        np.zeros(
            (len(f), len(terminal_component_modeler_data.modeler.matrix_indices_monitor)),
            dtype=complex,
        ),
        coords=coords,
    )
    b_sum = a_sum.copy()
    # Retrieve associated simulation data
    combined_directivity_data = None
    for port, amplitude in port_dict.items():
        if amplitude is not None:
            if np.isclose(amplitude, 0.0):
                continue
        sim_data_port = terminal_component_modeler_data.data[
            terminal_component_modeler_data.modeler.get_task_name(port)
        ]

        a, b = compute_wave_amplitudes_at_each_port(
            modeler=terminal_component_modeler_data.modeler,
            port_reference_impedances=terminal_component_modeler_data.port_reference_impedances,
            sim_data=sim_data_port,
            s_param_def="power",
        )
        # Select a possible subset of frequencies
        a = a.sel(f=f)
        b = b.sel(f=f)
        a_raw = a.sel(port=terminal_component_modeler_data.modeler.network_index(port))

        if amplitude is None:
            # No scaling performed when amplitude is None
            scaled_directivity_data = sim_data_port[rad_mon.name]
            scale_factor = 1.0
        else:
            scaled_directivity_data = (
                terminal_component_modeler_data._monitor_data_at_port_amplitude(
                    port, rad_mon.name, amplitude
                )
            )
            scale_factor = amplitude / a_raw
        a = scale_factor * a
        b = scale_factor * b

        # Combine the possibly scaled directivity data and the power wave amplitudes
        if combined_directivity_data is None:
            combined_directivity_data = scaled_directivity_data
        else:
            combined_directivity_data = combined_directivity_data + scaled_directivity_data
        a_sum += a
        b_sum += b

    # Compute and add power measures to results
    power_incident = np.real(0.5 * a_sum * np.conj(a_sum)).sum(dim="port")
    power_reflected = np.real(0.5 * b_sum * np.conj(b_sum)).sum(dim="port")
    return AntennaMetricsData.from_directivity_data(
        combined_directivity_data, power_incident, power_reflected
    )



from __future__ import annotations

from io import StringIO

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from tidy3d.components.viz import FLEXCOMPUTE_COLORS

MAX_POLES_TO_DISPLAY = 3


def summarize_material_library(matlib) -> str:
    """Return a user-friendly multiline string summarizing the entire material library."""
    lines = ["Material Library Summary:"]
    for key, mat_item in sorted(matlib.items()):
        if key != "graphene":
            lines.append(
                f"  - Key: {key}, Name: {mat_item.name}, "
                f"Default Variant: {mat_item.default}, "
                f"# Variants: {len(mat_item.variants)}"
            )
        else:
            lines.append(f"  - Key: {key}")
    return "\n".join(lines)


def summarize_material_item(m) -> str:
    """
    Return a concise summary of a single MaterialItem instance:
    includes name, default variant, plus the list of all variants.
    """
    lines = [f"Material Summary: {m.name}"]
    lines.append(f"  Default Variant: {m.default}")
    lines.append("  Variants:")
    for vname in m.variants.keys():
        lines.append(f"    - {vname}")
    return "\n".join(lines)


def variant_name(v):
    """Retrieve variant name for display."""
    default_name = "Unnamed Variant"
    name = default_name
    # Check attributes safely
    if hasattr(v, "medium") and hasattr(v.medium, "name"):
        name = v.medium.name
    elif hasattr(v, "ordinary") and hasattr(v.ordinary, "name"):
        name = v.ordinary.name
    # Ensure name is not None if getattr found the attr but it was None
    name = name or default_name

    return name


def summarize_medium(med) -> list[str]:
    """Returns relevant medium information for display."""
    lines = []

    if hasattr(med, "eps_inf"):
        lines.append(f"      eps_inf: {med.eps_inf}")
    if hasattr(med, "poles"):
        lines.append(f"      # poles: {len(med.poles)}")
        # optionally show some truncated details of each pole
        for i, (a, c) in enumerate(med.poles[:MAX_POLES_TO_DISPLAY]):
            lines.append(f"        - pole {i}: a={a}, c={c}")
        if len(med.poles) > MAX_POLES_TO_DISPLAY:
            lines.append("        ... (more poles omitted)")

    # or if MultiPhysicsMedium or other types, handle accordingly
    if hasattr(med, "frequency_range"):
        lines.append(f"      frequency_range: {med.frequency_range}")

    return lines


def summarize_variant_item(v) -> str:
    """
    Return a concise summary for a single VariantItem instance,
    highlighting references and main model parameters (e.g., eps_inf, # poles, freq. range).
    """
    name = variant_name(v)
    lines = []
    lines.append(f"Variant Summary: {name}")

    # references
    if v.reference:
        lines.append("  References:")
        for ref in v.reference:
            if ref.doi:
                lines.append(f"    DOI: {ref.doi}")
            elif ref.journal:
                lines.append(f"    Journal: {ref.journal}")
    else:
        lines.append("  References: None")

    lines.append(f"  data_url: {v.data_url or 'N/A'}")

    lines.append("  Medium model(s):")

    for medium_key, med in v.summarize_mediums.items():
        lines.append(f"    {medium_key}")

        lines += summarize_medium(med)

    return "\n".join(lines)


def repr_pretty_with_rich(obj, p, cycle):
    """Enable _repr_pretty_ for jupyter notebooks to use the rich printing output."""
    if cycle:
        p.text("MaterialLibrary(...)")
    else:
        # Render the __rich__ object to a string using a fake console
        sio = StringIO()
        console = Console(file=sio, force_terminal=True, width=100)
        console.print(obj.__rich__())
        p.text(sio.getvalue())


def add_medium_details_to_tree(medium, medium_node: Tree):
    """Adds details of a medium dictionary to a Rich Tree node."""

    if hasattr(medium, "eps_inf"):
        medium_node.add(f"[bold]eps_inf:[/bold] {medium.eps_inf}")
    if hasattr(medium, "poles"):
        num_poles = len(medium.poles)
        poles_node = medium_node.add(f"[bold]Poles: {num_poles}[/bold]")

        for i, (a, c) in enumerate(medium.poles[:MAX_POLES_TO_DISPLAY]):
            poles_node.add(f"pole {i}: a={a}, c={c}")
        if num_poles > MAX_POLES_TO_DISPLAY:
            poles_node.add("... (more poles omitted)")

    if hasattr(medium, "frequency_range"):
        medium_node.add(f"[bold]Freq. Range:[/bold] {medium.frequency_range}")


def summarize_material_library_rich(matlib):
    """Returns a Rich Table summarizing the MaterialLibrary."""

    table = Table(
        title=f"[bold {FLEXCOMPUTE_COLORS['brand_black']}]Material Library Summary[/]",
        show_header=True,
        header_style=f"bold {FLEXCOMPUTE_COLORS['brand_black']}",
        border_style=f"{FLEXCOMPUTE_COLORS['brand_black']}",
        expand=False,  # Don't force table to full terminal width
    )

    table.add_column("Key", style=f"{FLEXCOMPUTE_COLORS['brand_blue']}", width=18, justify="center")
    table.add_column(
        "Name", style=f"{FLEXCOMPUTE_COLORS['brand_purple']}", min_width=20, justify="center"
    )
    table.add_column(
        "Default Variant",
        style=f"{FLEXCOMPUTE_COLORS['brand_green']}",
        min_width=20,
        justify="center",
    )
    table.add_column("# Variants", style=f"{FLEXCOMPUTE_COLORS['brand_green']}", justify="center")

    # Iterate through the sorted items for consistent order
    for key, mat_item in sorted(matlib.items()):
        # Special handling for the 'graphene' key
        if key == "graphene":
            table.add_row(
                key,
                "Graphene",
                "---",  # Leave other cells blank for graphene
                "---",
            )
        else:
            name = mat_item.name
            default = mat_item.default if mat_item.default is not None else "[italic]None[/italic]"
            # Safely get length of variants (assuming it's a collection)
            num_variants = (
                str(len(mat_item.variants))
                if hasattr(mat_item, "variants") and mat_item.variants is not None
                else "[red]N/A[/red]"
            )

            table.add_row(key, name, default, num_variants)

    return table


def summarize_material_item_rich(m):
    """Returns a Rich Tree representation summarizing the MaterialItem."""
    tree_title = f"[bold {FLEXCOMPUTE_COLORS['brand_purple']}]Material Summary: {m.name}[/]"
    tree = Tree(tree_title)

    # Add a subtree for the variants
    variants_node = tree.add("[bold]Variants:[/bold]")
    if m.variants:
        # Sort keys for consistent display order
        sorted_variant_names = sorted(m.variants.keys())
        for vname in sorted_variant_names:
            # Add each variant name as a child node.
            variants_node.add(f"{vname}")
    else:
        # Indicate if there are no variants
        variants_node.add("[italic]None[/italic]")

    return Panel(tree, border_style=f"{FLEXCOMPUTE_COLORS['brand_purple']}", expand=False)


def summarize_variant_item_rich(v):
    """Returns Rich renderables for displaying a VariantItem summary."""

    name = variant_name(v)

    tree_title = Text(f"Variant Summary: {name}", style=f"bold {FLEXCOMPUTE_COLORS['brand_green']}")
    tree = Tree(tree_title)

    ref_node = tree.add("[bold]References[/bold]")
    if v.reference:
        for ref in v.reference:
            if hasattr(ref, "doi") and ref.doi:
                ref_node.add(f"[bold]DOI:[/bold] [link=https://doi.org/{ref.doi}]{ref.doi}[/link]")
            elif hasattr(ref, "journal") and ref.journal:
                ref_node.add(f"[bold]Journal:[/bold] {ref.journal}")
            else:
                ref_node.add("[italic]Reference details missing[/italic]")
    else:
        ref_node.add("[italic]None[/italic]")

    data_url_str = f"[bold]data_url:[/bold] {v.data_url or '[italic]N/A[/italic]'}"
    if (
        v.data_url
        and isinstance(v.data_url, str)
        and v.data_url.startswith(("http://", "https://"))
    ):
        data_url_str = f"[bold]data_url:[/bold] [link={v.data_url}]{v.data_url}[/link]"
    tree.add(data_url_str)

    mediums_node = tree.add("[bold]Medium model(s)[/bold]")

    mediums = v.summarize_mediums
    if mediums:
        for med_key, med in mediums.items():
            medium_type_node = mediums_node.add(
                f"[italic {FLEXCOMPUTE_COLORS['brand_blue']}]{med_key}[/]"
            )
            add_medium_details_to_tree(med, medium_type_node)
    else:
        mediums_node.add("[italic]None[/italic]")

    return Panel(tree, border_style=f"{FLEXCOMPUTE_COLORS['brand_green']}", expand=False)



"""Tool for generating an S matrix automatically from a Tidy3d simulation and terminal port definitions."""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.boundary import BroadbandModeABCSpec
from tidy3d.components.geometry.utils_2d import snap_coordinate_to_grid
from tidy3d.components.index import SimulationMap
from tidy3d.components.monitor import DirectivityMonitor
from tidy3d.components.simulation import Simulation
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.types import Ax, Complex
from tidy3d.components.viz import add_ax_if_none, equal_aspect
from tidy3d.constants import C_0, OHM
from tidy3d.exceptions import SetupError, Tidy3dKeyError, ValidationError
from tidy3d.log import log
from tidy3d.plugins.smatrix.component_modelers.base import (
    FWIDTH_FRAC,
    AbstractComponentModeler,
)
from tidy3d.plugins.smatrix.data.data_array import PortDataArray
from tidy3d.plugins.smatrix.ports.base_lumped import AbstractLumpedPort
from tidy3d.plugins.smatrix.ports.coaxial_lumped import CoaxialLumpedPort
from tidy3d.plugins.smatrix.ports.rectangular_lumped import LumpedPort
from tidy3d.plugins.smatrix.ports.types import TerminalPortType
from tidy3d.plugins.smatrix.ports.wave import WavePort
from tidy3d.plugins.smatrix.types import NetworkElement, NetworkIndex, SParamDef


class TerminalComponentModeler(AbstractComponentModeler):
    """
    Tool for modeling two-terminal multiport devices and computing port parameters
    with lumped and wave ports.


    Notes
    -----

    **References**

    .. [1]  R. B. Marks and D. F. Williams, "A general waveguide circuit theory,"
            J. Res. Natl. Inst. Stand. Technol., vol. 97, pp. 533, 1992.

    .. [2]  D. M. Pozar, Microwave Engineering, 4th ed. Hoboken, NJ, USA:
            John Wiley & Sons, 2012.
    """

    ports: tuple[TerminalPortType, ...] = pd.Field(
        (),
        title="Terminal Ports",
        description="Collection of lumped and wave ports associated with the network. "
        "For each port, one simulation will be run with a source that is associated with the port.",
    )

    run_only: Optional[tuple[NetworkIndex, ...]] = pd.Field(
        None,
        title="Run Only",
        description="Set of matrix indices that define the simulations to run. "
        "If ``None``, simulations will be run for all indices in the scattering matrix. "
        "If a tuple is given, simulations will be run only for the given matrix indices.",
    )

    element_mappings: tuple[tuple[NetworkElement, NetworkElement, Complex], ...] = pd.Field(
        (),
        title="Element Mappings",
        description="Tuple of S matrix element mappings, each described by a tuple of "
        "(input_element, output_element, coefficient), where the coefficient is the "
        "element_mapping coefficient describing the relationship between the input and output "
        "matrix element. If all elements of a given column of the scattering matrix are defined "
        "by ``element_mappings``, the simulation corresponding to this column is skipped automatically.",
    )

    radiation_monitors: tuple[DirectivityMonitor, ...] = pd.Field(
        (),
        title="Radiation Monitors",
        description="Facilitates the calculation of figures-of-merit for antennas. "
        "These monitor will be included in every simulation and record the radiated fields. ",
    )

    assume_ideal_excitation: bool = pd.Field(
        False,
        title="Assume Ideal Excitation",
        description="If ``True``, only the excited port is assumed to have a nonzero incident wave "
        "amplitude power. This choice simplifies the calculation of the scattering matrix. "
        "If ``False``, every entry in the vector of incident wave amplitudes (a) is calculated "
        "explicitly. This choice requires a matrix inversion when calculating the scattering "
        "matrix, but may lead to more accurate scattering parameters when there are "
        "reflections from simulation boundaries. ",
    )

    s_param_def: SParamDef = pd.Field(
        "pseudo",
        title="Scattering Parameter Definition",
        description="Whether to compute scattering parameters using the 'pseudo' or 'power' wave definitions.",
    )

    @pd.root_validator(pre=False)
    def _warn_rf_license(cls, values):
        log.warning(
            "ℹ️ ⚠️ RF simulations are subject to new license requirements in the future. You have instantiated at least one RF-specific component.",
            log_once=True,
        )
        return values

    @pd.root_validator(pre=False)
    def _warn_refactor_2_10(cls, values):
        log.warning(
            f"ℹ️ ⚠️ The {cls.__name__} class was refactored in tidy3d version 2.10. Migration documentation will be provided, and existing functionality can be accessed in a different way.",
            log_once=True,
        )
        return values

    @property
    def _sim_with_sources(self) -> Simulation:
        """Instance of :class:`.Simulation` with all sources and absorbers added for each port, for troubleshooting."""

        sources = [port.to_source(self._source_time) for port in self.ports]
        absorbers = [
            port.to_absorber()
            for port in self.ports
            if isinstance(port, WavePort) and port.absorber
        ]
        return self.simulation.updated_copy(sources=sources, internal_absorbers=absorbers)

    @equal_aspect
    @add_ax_if_none
    def plot_sim(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plot a :class:`.Simulation` with all sources and absorbers.

        This is a convenience method to visualize the simulation setup for
        troubleshooting. It shows all sources and absorbers for each port.

        Parameters
        ----------
        x : float, optional
            x-coordinate for the cross-section.
        y : float, optional
            y-coordinate for the cross-section.
        z : float, optional
            z-coordinate for the cross-section.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on.
        **kwargs
            Keyword arguments passed to :meth:`.Simulation.plot`.

        Returns
        -------
        matplotlib.axes.Axes
            The axes with the plot.
        """
        return self._sim_with_sources.plot(x=x, y=y, z=z, ax=ax, **kwargs)

    @equal_aspect
    @add_ax_if_none
    def plot_sim_eps(
        self,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        ax: Ax = None,
        **kwargs,
    ) -> Ax:
        """Plot permittivity of the :class:`.Simulation`.

        This method shows the permittivity distribution of the simulation with
        all sources and absorbers added for each port.

        Parameters
        ----------
        x : float, optional
            x-coordinate for the cross-section.
        y : float, optional
            y-coordinate for the cross-section.
        z : float, optional
            z-coordinate for the cross-section.
        ax : matplotlib.axes.Axes, optional
            Axes to plot on.
        **kwargs
            Keyword arguments passed to :meth:`.Simulation.plot_eps`.

        Returns
        -------
        matplotlib.axes.Axes
            The axes with the plot.
        """

        return self._sim_with_sources.plot_eps(x=x, y=y, z=z, ax=ax, **kwargs)

    @staticmethod
    def network_index(port: TerminalPortType, mode_index: Optional[int] = None) -> NetworkIndex:
        """Converts the port, and a ``mode_index`` when the port is a :class:`.WavePort`, to a unique string specifier.

        Parameters
        ----------
        port : ``TerminalPortType``
            The port to convert to an index.
        mode_index : Optional[int]
            Selects a single mode from those supported by the ``port``, which is only used when
            the ``port`` is a :class:`.WavePort`

        Returns
        -------
        NetworkIndex
            A unique string that is used to identify the row/column of the scattering matrix.
        """
        # Currently the mode_index is ignored, but will be supported once multimodal WavePorts are enabled.
        return f"{port.name}"

    @cached_property
    def network_dict(self) -> dict[NetworkIndex, tuple[TerminalPortType, int]]:
        """Dictionary associating each unique ``NetworkIndex`` to a port and mode index."""
        network_dict = {}
        for port in self.ports:
            mode_index = None
            if isinstance(port, WavePort):
                mode_index = port.mode_index
            key = TerminalComponentModeler.network_index(port, mode_index)
            network_dict[key] = (port, mode_index)
        return network_dict

    @cached_property
    def matrix_indices_monitor(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the possible matrix indices."""
        matrix_indices = []
        for port in self.ports:
            if isinstance(port, WavePort):
                matrix_indices.append(self.network_index(port, port.mode_index))
            else:
                matrix_indices.append(self.network_index(port))
        return tuple(matrix_indices)

    @cached_property
    def matrix_indices_source(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the source matrix indices, which may be less than the total number of
        ports."""
        return super().matrix_indices_source

    @cached_property
    def matrix_indices_run_sim(self) -> tuple[NetworkIndex, ...]:
        """Tuple of all the matrix indices that will be used to run simulations."""
        return super().matrix_indices_run_sim

    @cached_property
    def sim_dict(self) -> SimulationMap:
        """Generate all the :class:`.Simulation` objects for the port parameter calculation."""

        sim_dict = {}
        # Now, create simulations with wave port sources and mode solver monitors for computing port modes
        for network_index in self.matrix_indices_run_sim:
            task_name, sim_with_src = self._add_source_to_sim(network_index)
            sim_dict[task_name] = sim_with_src

        # Check final simulations for grid size at ports
        for _, sim in sim_dict.items():
            TerminalComponentModeler._check_grid_size_at_ports(sim, self._lumped_ports)
            TerminalComponentModeler._check_grid_size_at_wave_ports(sim, self._wave_ports)

        return SimulationMap(keys=tuple(sim_dict.keys()), values=tuple(sim_dict.values()))

    @cached_property
    def base_sim(self) -> Simulation:
        """The base simulation with all grid refinement options, port loads (if present), and monitors added,
        which is only missing the source excitations.
        """
        # internal mesh override and snapping points are automatically generated from lumped elements.
        lumped_resistors = [port.to_load() for port in self._lumped_ports]

        # Apply the highest frequency in the simulation to define the grid, rather than the
        # source's central frequency, to ensure an accurate solution over the entire range
        grid_spec = self.simulation.grid_spec.copy(
            update={
                "wavelength": C_0 / np.max(self.freqs),
            }
        )

        # Make an initial simulation with new grid_spec to determine where LumpedPorts are snapped
        sim_wo_source = self.simulation.updated_copy(
            grid_spec=grid_spec, lumped_elements=lumped_resistors
        )
        snap_centers = {}
        for port in self._lumped_ports:
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                sim_wo_source.grid, port_center_on_axis, port.injection_axis
            )
            snap_centers[port.name] = new_port_center

        # Create monitors and snap to the center positions
        field_monitors = [
            mon
            for port in self.ports
            for mon in port.to_monitors(
                self.freqs, snap_center=snap_centers.get(port.name), grid=sim_wo_source.grid
            )
        ]

        new_mnts = list(self.simulation.monitors) + field_monitors

        if self.radiation_monitors is not None:
            new_mnts = new_mnts + list(self.radiation_monitors)

        new_lumped_elements = list(self.simulation.lumped_elements) + [
            port.to_load(snap_center=snap_centers[port.name]) for port in self._lumped_ports
        ]

        # Add mesh overrides for any wave ports present
        mesh_overrides = list(sim_wo_source.grid_spec.override_structures)
        for wave_port in self._wave_ports:
            if wave_port.num_grid_cells is not None:
                mesh_overrides.extend(wave_port.to_mesh_overrides())
        new_grid_spec = sim_wo_source.grid_spec.updated_copy(override_structures=mesh_overrides)

        new_absorbers = list(sim_wo_source.internal_absorbers)
        for wave_port in self._wave_ports:
            if wave_port.absorber:
                # absorbers are shifted together with sources
                mode_src_pos = wave_port.center[
                    wave_port.injection_axis
                ] + self._shift_value_signed(wave_port)
                port_absorber = wave_port.to_absorber(
                    snap_center=mode_src_pos,
                    freq_spec=BroadbandModeABCSpec(
                        frequency_range=(np.min(self.freqs), np.max(self.freqs))
                    ),
                )
                new_absorbers.append(port_absorber)

        update_dict = {
            "monitors": new_mnts,
            "lumped_elements": new_lumped_elements,
            "grid_spec": new_grid_spec,
            "internal_absorbers": new_absorbers,
        }

        # This is the new default simulation will all shared components added
        return sim_wo_source.copy(update=update_dict)

    def _add_source_to_sim(self, source_index: NetworkIndex) -> tuple[str, Simulation]:
        """Adds the source corresponding to the ``source_index`` to the base simulation."""
        port, mode_index = self.network_dict[source_index]
        if isinstance(port, WavePort):
            # Source is placed just before the field monitor of the port
            mode_src_pos = port.center[port.injection_axis] + self._shift_value_signed(port)
            port_source = port.to_source(self._source_time, snap_center=mode_src_pos)
        else:
            port_center_on_axis = port.center[port.injection_axis]
            new_port_center = snap_coordinate_to_grid(
                self.base_sim.grid, port_center_on_axis, port.injection_axis
            )
            port_source = port.to_source(
                self._source_time, snap_center=new_port_center, grid=self.base_sim.grid
            )
        task_name = self.get_task_name(port=port, mode_index=mode_index)
        return (task_name, self.base_sim.updated_copy(sources=[port_source]))

    @cached_property
    def _source_time(self):
        """Helper to create a time domain pulse for the frequency range of interest."""
        if len(self.freqs) == 1:
            freq0 = self.freqs[0]
            return GaussianPulse(freq0=self.freqs[0], fwidth=freq0 * FWIDTH_FRAC)

        # Using the minimum_source_bandwidth, ensure we don't create a pulse that is too narrowband
        # when fmin and fmax are close together
        return GaussianPulse.from_frequency_range(
            fmin=np.min(self.freqs),
            fmax=np.max(self.freqs),
            remove_dc_component=self.remove_dc_component,
            minimum_source_bandwidth=FWIDTH_FRAC,
        )

    @pd.validator("simulation")
    def _validate_3d_simulation(cls, val):
        """Error if :class:`.Simulation` is not a 3D simulation"""

        if val.size.count(0.0) > 0:
            raise ValidationError(
                f"'{cls.__name__}' must be setup with a 3D simulation with all sizes greater than 0."
            )
        return val

    @pd.validator("radiation_monitors")
    def _validate_radiation_monitors(cls, val, values):
        freqs = set(values.get("freqs"))
        for rad_mon in val:
            mon_freqs = rad_mon.freqs
            is_subset = freqs.issuperset(mon_freqs)
            if not is_subset:
                raise ValidationError(
                    f"The frequencies in the radiation monitor '{rad_mon.name}' "
                    f"must be equal to or a subset of the frequencies in the '{cls.__name__}'."
                )
        return val

    @staticmethod
    def _check_grid_size_at_ports(
        simulation: Simulation, ports: list[Union[LumpedPort, CoaxialLumpedPort]]
    ):
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        yee_grid = simulation.grid.yee
        for port in ports:
            port._check_grid_size(yee_grid)

    @staticmethod
    def _check_grid_size_at_wave_ports(simulation: Simulation, ports: list[WavePort]):
        """Raises :class:`.SetupError` if the grid is too coarse at port locations"""
        for port in ports:
            disc_grid = simulation.discretize(port)
            check_axes = port.transverse_axes
            msg_header = f"'WavePort' '{port.name}' "
            for axis in check_axes:
                sim_size = simulation.size[axis]
                dim_cells = disc_grid.num_cells[axis]
                if sim_size > 0 and dim_cells <= 2:
                    small_dim = "xyz"[axis]
                    raise SetupError(
                        msg_header + f"is too small along the "
                        f"'{small_dim}' axis. Less than '3' grid cells were detected. "
                        "Please ensure that the port's 'num_grid_cells' is not 'None'. "
                        "You also may need to use an 'AutoGrid' or `QuasiUniformGrid` "
                        "for the simulation passed to the 'TerminalComponentModeler'."
                    )

    @cached_property
    def _lumped_ports(self) -> list[AbstractLumpedPort]:
        """A list of all lumped ports in the :class:`.TerminalComponentModeler`"""
        return [port for port in self.ports if isinstance(port, AbstractLumpedPort)]

    @cached_property
    def _wave_ports(self) -> list[WavePort]:
        """A list of all wave ports in the :class:`.TerminalComponentModeler`"""
        return [port for port in self.ports if isinstance(port, WavePort)]

    @staticmethod
    def _set_port_data_array_attributes(data_array: PortDataArray) -> PortDataArray:
        """Helper to set additional metadata for ``PortDataArray``."""
        data_array.name = "Z0"
        return data_array.assign_attrs(units=OHM, long_name="characteristic impedance")

    def get_radiation_monitor_by_name(self, monitor_name: str) -> DirectivityMonitor:
        """Find and return a :class:`.DirectivityMonitor` monitor by its name.

        Parameters
        ----------
        monitor_name : str
            Name of the monitor to find.

        Returns
        -------
        :class:`.DirectivityMonitor`
            The monitor matching the given name.

        Raises
        ------
        ``Tidy3dKeyError``
            If no monitor with the given name exists.
        """
        for monitor in self.radiation_monitors:
            if monitor.name == monitor_name:
                return monitor
        raise Tidy3dKeyError(f"No radiation monitor named '{monitor_name}'.")


TerminalComponentModeler.update_forward_refs()



"""Utilities for autograd derivative computation and field gradient evaluation."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import Callable, Optional, Union

import numpy as np
import xarray as xr

from tidy3d.components.data.data_array import FreqDataArray, ScalarFieldDataArray
from tidy3d.components.types import ArrayLike, Bound, tidycomplex
from tidy3d.constants import C_0, EPSILON_0, LARGE_NUMBER, MU_0
from tidy3d.log import log

from .constants import (
    DEFAULT_WAVELENGTH_FRACTION,
    GRADIENT_DTYPE_COMPLEX,
    GRADIENT_DTYPE_FLOAT,
    MINIMUM_SPACING_FRACTION,
)
from .types import PathType
from .utils import get_static

FieldData = dict[str, ScalarFieldDataArray]
PermittivityData = dict[str, ScalarFieldDataArray]
EpsType = Union[tidycomplex, FreqDataArray]


class LazyInterpolator:
    """Lazy wrapper for interpolators that creates them on first access."""

    def __init__(self, creator_func: Callable):
        """Initialize with a function that creates the interpolator when called."""
        self.creator_func = creator_func
        self._interpolator = None

    def __call__(self, *args, **kwargs):
        """Create interpolator on first call and delegate to it."""
        if self._interpolator is None:
            self._interpolator = self.creator_func()
        return self._interpolator(*args, **kwargs)


@dataclass
class DerivativeInfo:
    """Stores derivative information passed to the ``._compute_derivatives`` methods.

    This dataclass contains all the field data and parameters needed for computing
    gradients with respect to geometry perturbations.
    """

    # Required fields
    paths: list[PathType]
    """List of paths to the traced fields that need derivatives calculated."""

    E_der_map: FieldData
    """Electric field gradient map.
    Dataset where the field components ("Ex", "Ey", "Ez") store the multiplication
    of the forward and adjoint electric fields. The tangential components of this
    dataset are used when computing adjoint gradients for shifting boundaries.
    All components are used when computing volume-based gradients."""

    D_der_map: FieldData
    """Displacement field gradient map.
    Dataset where the field components ("Ex", "Ey", "Ez") store the multiplication
    of the forward and adjoint displacement fields. The normal component of this
    dataset is used when computing adjoint gradients for shifting boundaries."""

    E_fwd: FieldData
    """Forward electric fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the forward
    electric fields used for computing gradients for a given structure."""

    E_adj: FieldData
    """Adjoint electric fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the adjoint
    electric fields used for computing gradients for a given structure."""

    D_fwd: FieldData
    """Forward displacement fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the forward
    displacement fields used for computing gradients for a given structure."""

    D_adj: FieldData
    """Adjoint displacement fields.
    Dataset where the field components ("Ex", "Ey", "Ez") represent the adjoint
    displacement fields used for computing gradients for a given structure."""

    eps_data: PermittivityData
    """Permittivity dataset.
    Dataset of relative permittivity values along all three dimensions.
    Used for automatically computing permittivity inside or outside of a simple geometry."""

    eps_in: EpsType
    """Permittivity inside the Structure.
    Typically computed from Structure.medium.eps_model.
    Used when it cannot be computed from eps_data or when eps_approx=True."""

    eps_out: EpsType
    """Permittivity outside the Structure.
    Typically computed from Simulation.medium.eps_model.
    Used when it cannot be computed from eps_data or when eps_approx=True."""

    bounds: Bound
    """Geometry bounds.
    Bounds corresponding to the structure, used in Medium calculations."""

    bounds_intersect: Bound
    """Geometry and simulation intersection bounds.
    Bounds corresponding to the minimum intersection between the structure
    and the simulation it is contained in."""

    simulation_bounds: Bound
    """Simulation bounds.
    Bounds corresponding to the simulation domain containing this structure.
    Unlike bounds_intersect, this is independent of the structure's bounds and
    is purely based on the simulation geometry."""

    frequencies: ArrayLike
    """Frequencies at which the adjoint gradient should be computed."""

    H_der_map: Optional[FieldData] = None
    """Magnetic field gradient map.
    Dataset where the field components ("Hx", "Hy", "Hz") store the multiplication
    of the forward and adjoint magnetic fields. The tangential component of this
    dataset is used when computing adjoint gradients for shifting boundaries of
    structures composed of PEC mediums."""

    H_fwd: Optional[FieldData] = None
    """Forward magnetic fields.
    Dataset where the field components ("Hx", "Hy", "Hz") represent the forward
    magnetic fields used for computing gradients for a given structure."""

    H_adj: Optional[FieldData] = None
    """Adjoint magnetic fields.
    Dataset where the field components ("Hx", "Hy", "Hz") represent the adjoint
    magnetic fields used for computing gradients for a given structure."""

    # Optional fields with defaults
    eps_background: Optional[EpsType] = None
    """Permittivity in background.
    Permittivity outside of the Structure as manually specified by
    Structure.background_medium."""

    eps_no_structure: Optional[ScalarFieldDataArray] = None
    """Permittivity without structure.
    The permittivity of the original simulation without the structure that is
    being differentiated with respect to. Used to approximate permittivity
    outside of the structure for shape optimization."""

    eps_inf_structure: Optional[ScalarFieldDataArray] = None
    """Permittivity with infinite structure.
    The permittivity of the original simulation where the structure being
    differentiated with respect to is infinitely large. Used to approximate
    permittivity inside of the structure for shape optimization."""

    eps_approx: bool = False
    """Use permittivity approximation.
    If True, approximates outside permittivity using Simulation.medium and
    the inside permittivity using Structure.medium. Only set True for
    GeometryGroup handling where it is difficult to automatically evaluate
    the inside and outside relative permittivity for each geometry."""

    is_medium_pec: bool = False
    """Indicates if structure material is PEC.
    If True, the structure contains a PEC material which changes the gradient
    formulation at the boundary compared to the dielectric case."""

    interpolators: Optional[dict] = None
    """Pre-computed interpolators.
    Optional pre-computed interpolators for field components and permittivity data.
    When provided, avoids redundant interpolator creation for multiple geometries
    sharing the same field data. This significantly improves performance for
    GeometryGroup processing."""

    # private cache for interpolators
    _interpolators_cache: dict = field(default_factory=dict, init=False, repr=False)

    def updated_copy(self, **kwargs):
        """Create a copy with updated fields."""
        kwargs.pop("deep", None)
        kwargs.pop("validate", None)
        return replace(self, **kwargs)

    @staticmethod
    def _nan_to_num_if_needed(coords: np.ndarray) -> np.ndarray:
        """Convert NaN and infinite values to finite numbers, optimized for finite inputs."""
        # skip check for small arrays
        if coords.size < 1000:
            return np.nan_to_num(coords, posinf=LARGE_NUMBER, neginf=-LARGE_NUMBER)

        if np.isfinite(coords).all():
            return coords
        return np.nan_to_num(coords, posinf=LARGE_NUMBER, neginf=-LARGE_NUMBER)

    @staticmethod
    def _evaluate_with_interpolators(
        interpolators: dict, coords: np.ndarray
    ) -> dict[str, np.ndarray]:
        """Evaluate field components at coordinates using cached interpolators.

        Parameters
        ----------
        interpolators : dict
            Dictionary mapping field component names to ``RegularGridInterpolator`` objects.
        coords : np.ndarray
            Spatial coordinates (N, 3) where fields are evaluated.

        Returns
        -------
        dict[str, np.ndarray]
            Dictionary mapping component names to field values at coordinates.
        """
        coords = DerivativeInfo._nan_to_num_if_needed(coords)
        if coords.dtype != GRADIENT_DTYPE_FLOAT and coords.dtype != GRADIENT_DTYPE_COMPLEX:
            coords = coords.astype(GRADIENT_DTYPE_FLOAT, copy=False)
        return {name: interp(coords) for name, interp in interpolators.items()}

    def create_interpolators(self, dtype=GRADIENT_DTYPE_FLOAT) -> dict:
        """Create interpolators for field components and permittivity data.

        Creates and caches ``RegularGridInterpolator`` objects for all field components
        (E_fwd, E_adj, D_fwd, D_adj) and permittivity data (eps_inf, eps_no).
        This caching strategy significantly improves performance by avoiding
        repeated interpolator construction in gradient evaluation loops.

        Parameters
        ----------
        dtype : np.dtype = GRADIENT_DTYPE_FLOAT
            Data type for interpolation coordinates and values.

        Returns
        -------
        dict
            Nested dictionary structure:
            - Field data: {"E_fwd": {"Ex": interpolator, ...}, ...}
            - Permittivity: {"eps_inf": interpolator, "eps_no": interpolator}
        """
        from scipy.interpolate import RegularGridInterpolator

        cache_key = str(dtype)
        if cache_key in self._interpolators_cache:
            return self._interpolators_cache[cache_key]

        interpolators = {}
        coord_cache = {}

        def _make_lazy_interpolator_group(field_data_dict, group_key, is_field_group=True):
            """Helper to create a group of lazy interpolators."""
            if is_field_group:
                interpolators[group_key] = {}

            for component_name, arr in field_data_dict.items():
                # use object ID for caching to handle shared grids
                arr_id = id(arr.data)
                if arr_id not in coord_cache:
                    points = tuple(c.data.astype(dtype, copy=False) for c in (arr.x, arr.y, arr.z))
                    coord_cache[arr_id] = points
                points = coord_cache[arr_id]

                def creator_func(arr=arr, points=points):
                    data = arr.data.astype(
                        GRADIENT_DTYPE_COMPLEX if np.iscomplexobj(arr.data) else dtype, copy=False
                    )
                    # create interpolator with frequency dimension
                    if "f" in arr.dims:
                        freq_coords = arr.coords["f"].data.astype(dtype, copy=False)
                        # ensure frequency dimension is last
                        if arr.dims != ("x", "y", "z", "f"):
                            freq_dim_idx = arr.dims.index("f")
                            axes = list(range(data.ndim))
                            axes.append(axes.pop(freq_dim_idx))
                            data = np.transpose(data, axes)
                    else:
                        # single frequency case - add singleton dimension
                        freq_coords = np.array([0.0], dtype=dtype)
                        data = data[..., np.newaxis]

                    points_with_freq = (*points, freq_coords)
                    # If PEC, use nearest interpolation instead of linear to avoid interpolating
                    # with field values inside the PEC (which are 0). Instead, we make sure to
                    # choose interplation points such that their nearest location is outside of
                    # the PEC surface.
                    method = "nearest" if self.is_medium_pec else "linear"
                    interpolator_obj = RegularGridInterpolator(
                        points_with_freq, data, method=method, bounds_error=False, fill_value=None
                    )

                    def interpolator(coords):
                        # coords: (N, 3) spatial points
                        n_points = coords.shape[0]
                        n_freqs = len(freq_coords)

                        # build coordinates with frequency dimension
                        coords_with_freq = np.empty((n_points * n_freqs, 4), dtype=coords.dtype)
                        coords_with_freq[:, :3] = np.repeat(coords, n_freqs, axis=0)
                        coords_with_freq[:, 3] = np.tile(freq_coords, n_points)

                        result = interpolator_obj(coords_with_freq)
                        return result.reshape(n_points, n_freqs)

                    return interpolator

                if is_field_group:
                    interpolators[group_key][component_name] = LazyInterpolator(creator_func)
                else:
                    interpolators[component_name] = LazyInterpolator(creator_func)

        # process field interpolators (nested dictionaries)
        interpolator_groups = [
            ("E_fwd", self.E_fwd),
            ("E_adj", self.E_adj),
            ("D_fwd", self.D_fwd),
            ("D_adj", self.D_adj),
        ]
        if self.is_medium_pec:
            interpolator_groups += [("H_fwd", self.H_fwd), ("H_adj", self.H_adj)]
        for group_key, data_dict in interpolator_groups:
            _make_lazy_interpolator_group(data_dict, group_key, is_field_group=True)

        if self.eps_inf_structure is not None:
            _make_lazy_interpolator_group(
                {"eps_inf": self.eps_inf_structure}, None, is_field_group=False
            )
        if self.eps_no_structure is not None:
            _make_lazy_interpolator_group(
                {"eps_no": self.eps_no_structure}, None, is_field_group=False
            )

        self._interpolators_cache[cache_key] = interpolators
        return interpolators

    def evaluate_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: Optional[dict] = None,
    ) -> np.ndarray:
        """Compute adjoint gradients at surface points for shape optimization.

        Implements the surface integral formulation for computing gradients with respect
        to geometry perturbations.

        Parameters
        ----------
        spatial_coords : np.ndarray
            (N, 3) array of surface evaluation points.
        normals : np.ndarray
            (N, 3) array of outward-pointing normal vectors at each surface point.
        perps1 : np.ndarray
            (N, 3) array of first tangent vectors perpendicular to normals.
        perps2 : np.ndarray
            (N, 3) array of second tangent vectors perpendicular to both normals and perps1.
        interpolators : dict = None
            Pre-computed field interpolators for efficiency.

        Returns
        -------
        np.ndarray
            (N,) array of gradient values at each surface point. Must be integrated
            with appropriate quadrature weights to get total gradient.
        """
        if interpolators is None:
            raise NotImplementedError(
                "Direct field evaluation without interpolators is not implemented. "
                "Please create interpolators using 'create_interpolators()' first."
            )

        if "eps_no" in interpolators:
            eps_out = interpolators["eps_no"](spatial_coords)
        else:
            # use eps_background if available, otherwise use eps_out
            eps_to_prepare = (
                self.eps_background if self.eps_background is not None else self.eps_out
            )
            eps_out = self._prepare_epsilon(eps_to_prepare)

        if self.is_medium_pec:
            vjps = self._evaluate_pec_gradient_at_points(
                spatial_coords, normals, perps1, perps2, interpolators, eps_out
            )
        else:
            vjps = self._evaluate_dielectric_gradient_at_points(
                spatial_coords, normals, perps1, perps2, interpolators, eps_out
            )

        # sum over frequency dimension
        vjps = np.sum(vjps, axis=-1)

        return vjps

    def _evaluate_dielectric_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: dict,
        # todo: type
        eps_out,
    ) -> np.ndarray:
        # evaluate all field components at surface points
        E_fwd_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["E_fwd"].items()
        }
        E_adj_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["E_adj"].items()
        }
        D_fwd_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["D_fwd"].items()
        }
        D_adj_at_coords = {
            name: interp(spatial_coords) for name, interp in interpolators["D_adj"].items()
        }

        if "eps_inf" in interpolators:
            eps_in = interpolators["eps_inf"](spatial_coords)
        else:
            eps_in = self._prepare_epsilon(self.eps_in)

        delta_eps_inv = 1.0 / eps_in - 1.0 / eps_out
        delta_eps = eps_in - eps_out

        # project fields onto local surface basis (normal + two tangents)
        D_fwd_norm = self._project_in_basis(D_fwd_at_coords, basis_vector=normals)
        D_adj_norm = self._project_in_basis(D_adj_at_coords, basis_vector=normals)

        E_fwd_perp1 = self._project_in_basis(E_fwd_at_coords, basis_vector=perps1)
        E_adj_perp1 = self._project_in_basis(E_adj_at_coords, basis_vector=perps1)

        E_fwd_perp2 = self._project_in_basis(E_fwd_at_coords, basis_vector=perps2)
        E_adj_perp2 = self._project_in_basis(E_adj_at_coords, basis_vector=perps2)

        D_der_norm = D_fwd_norm * D_adj_norm
        E_der_perp1 = E_fwd_perp1 * E_adj_perp1
        E_der_perp2 = E_fwd_perp2 * E_adj_perp2

        vjps = -delta_eps_inv * D_der_norm + E_der_perp1 * delta_eps + E_der_perp2 * delta_eps

        return vjps

    def _evaluate_pec_gradient_at_points(
        self,
        spatial_coords: np.ndarray,
        normals: np.ndarray,
        perps1: np.ndarray,
        perps2: np.ndarray,
        interpolators: dict,
        # todo: type
        eps_out,
    ) -> np.ndarray:
        def _adjust_spatial_coords_pec(grid_centers: dict[str, np.ndarray]):
            """Assuming a nearest interpolation, adjust the interpolation points given the grid
            defined by `grid_centers` and using `spatial_coords` as a starting point such that we
            select a point outside of the PEC boundary.

                 *** (nearest point outside boundary)
                  ^
                  | n (normal direction)
                  |
            _.-~'`-._.-~'`-._ (PEC surface)
                  * (nearest point)

            Parameters
            ----------
            grid_centers: dict[str, np.ndarray]
                The grid points for a given field component indexed by dimension. These grid points
                are used to find the nearest snapping point and adjust the inerpolation coordinates
                to ensure we fall outside of the PEC surface.

            Returns
            -------
            (np.ndarray, np.ndarray)
                (N, 3) array of coordinate centers at which to interpolate such that they line up
                with a grid center and are outside the PEC surface
                (N,) array of distances from the nearest interpolation points to the desired surface
                edge points specified by `spatial_coords`

            """
            grid_ddim = np.zeros_like(normals)
            for idx, dim in enumerate("xyz"):
                expanded_coords = np.expand_dims(spatial_coords[:, idx], axis=1)
                grid_centers_select = grid_centers[dim]

                diff = np.abs(expanded_coords - grid_centers_select)

                nearest_grid = np.argmin(diff, axis=-1)
                nearest_grid = np.minimum(np.maximum(nearest_grid, 1), len(grid_centers_select) - 1)

                # compute the local grid spacing near the boundary
                grid_ddim[:, idx] = (
                    grid_centers_select[nearest_grid] - grid_centers_select[nearest_grid - 1]
                )

            # assuming we move in the normal direction, finds which dimension we need to move the least
            # in order to ensure we snap to a point outside the boundary in the worst case (i.e. - the
            # nearest point is just inside the surface)
            min_movement_index = np.argmin(
                np.abs(grid_ddim) / (np.abs(normals) + np.finfo(normals.dtype).min), axis=1
            )

            selection = (np.arange(normals.shape[0]), min_movement_index)
            coords_dn = np.expand_dims(np.abs(grid_ddim[selection]), axis=1)

            # adjust coordinates by half a grid point outside boundary such that nearest interpolation
            # point snaps to outside the boundary
            adjust_spatial_coords = spatial_coords + normals * 0.5 * coords_dn

            edge_distance = np.zeros_like(adjust_spatial_coords[:, 0])
            for idx, dim in enumerate("xyz"):
                expanded_adjusted_coords = np.expand_dims(adjust_spatial_coords[:, idx], axis=1)
                grid_centers_select = grid_centers[dim]

                # find nearest grid point from the adjusted coordinates
                diff = np.abs(expanded_adjusted_coords - grid_centers_select)
                nearest_grid = np.argmin(diff, axis=-1)

                # compute edge distance from the nearest interpolated point to the boundary edge
                edge_distance += (
                    np.abs(spatial_coords[:, idx] - grid_centers_select[nearest_grid]) ** 2
                )

            # this edge distance is useful when correcting for edge singularities from the PEC material
            # and is used when the PEC PolySlab structure has zero thickness
            edge_distance = np.sqrt(edge_distance)

            return adjust_spatial_coords, edge_distance

        def _snap_coordinate_outside(field_components: FieldData):
            """Helper function to perform coordinate adjustment and compute edge distance for each
            component in `field_components`.

            Parameters
            ----------
            field_components: FieldData
                The field components (i.e - Ex, Ey, Ez, Hx, Hy, Hz) that we would like to sample just
                outside the PEC surface using nearest interpolation.

            Returns
            -------
            dict[str, dict[str, np.ndarray]]
                Dictionary mapping each field component name to a dictionary of adjusted coordinates
                and edge distances for that component.
            """
            adjustment = {}
            for name in field_components:
                field_component = field_components[name]
                field_component_coords = field_component.coords

                adjusted_coords, edge_distance = _adjust_spatial_coords_pec(
                    {
                        key: np.array(field_component_coords[key].values)
                        for key in field_component_coords
                    }
                )
                adjustment[name] = {"coords": adjusted_coords, "edge_distance": edge_distance}

            return adjustment

        def _interpolate_field_components(interp_coords, field_name):
            return {
                name: interp(interp_coords[name]["coords"])
                for name, interp in interpolators[field_name].items()
            }

        # adjust coordinates for PEC to be outside structure bounds and get edge distance for singularity correction.
        E_fwd_coords_adjusted = _snap_coordinate_outside(self.E_fwd)
        E_adj_coords_adjusted = _snap_coordinate_outside(self.E_adj)

        H_fwd_coords_adjusted = _snap_coordinate_outside(self.H_fwd)
        H_adj_coords_adjusted = _snap_coordinate_outside(self.H_adj)

        # using the adjusted coordinates, evaluate all field components at surface points
        E_fwd_at_coords = _interpolate_field_components(E_fwd_coords_adjusted, "E_fwd")
        E_adj_at_coords = _interpolate_field_components(E_adj_coords_adjusted, "E_adj")
        H_fwd_at_coords = _interpolate_field_components(H_fwd_coords_adjusted, "H_fwd")
        H_adj_at_coords = _interpolate_field_components(H_adj_coords_adjusted, "H_adj")

        structure_sizes = np.array(
            [self.bounds[1][idx] - self.bounds[0][idx] for idx in range(len(self.bounds[0]))]
        )

        is_flat_perp_dim1 = np.isclose(np.abs(np.sum(perps1[0] * structure_sizes)), 0.0)
        is_flat_perp_dim2 = np.isclose(np.abs(np.sum(perps2[0] * structure_sizes)), 0.0)
        flat_perp_dims = [is_flat_perp_dim1, is_flat_perp_dim2]

        # check if this integration is happening along an edge in which case we will eliminate
        # on of the H field integration components and apply singularity correction
        pec_line_integration = is_flat_perp_dim1 or is_flat_perp_dim2

        def _compute_singularity_correction(adjustment_: dict[str, dict[str, np.ndarray]]):
            """
            Given the `adjustment_` which contains the distance from the PEC edge each field
            component is nearest interpolated at, computes the singularity correction when
            working with 2D PEC using the average edge_distance for each component. In the case
            of 3D PEC gradients, no singularity correction is applied so an array of ones is returned.

            Parameters
            ----------
            adjustment_: dict[str, dict[str, np.ndarray]]
                Dictionary that maps field component name to a dictionary containing the coordinate
                adjustment and the distance to the PEC edge for those coordinates. The edge distance
                is used for 2D PEC singularity correction.

            Returns
            -------
            np.ndarray
                Returns the singularity correction which has shape (N,) where there are N points in
                `spatial_coords`
            """
            return (
                (
                    0.5
                    * np.pi
                    * np.mean([adjustment_[name]["edge_distance"] for name in adjustment_], axis=0)
                )
                if pec_line_integration
                else np.ones_like(spatial_coords, shape=spatial_coords.shape[0])
            )

        E_norm_singularity_correction = np.expand_dims(
            _compute_singularity_correction(E_fwd_coords_adjusted), axis=1
        )
        H_perp_singularity_correction = np.expand_dims(
            _compute_singularity_correction(H_fwd_coords_adjusted), axis=1
        )

        E_fwd_norm = self._project_in_basis(E_fwd_at_coords, basis_vector=normals)
        E_adj_norm = self._project_in_basis(E_adj_at_coords, basis_vector=normals)

        # compute the normal E contribution to the gradient (the tangential E contribution
        # is 0 in the case of PEC since this field component is continuous and thus 0 at
        # the boundary)
        contrib_E = E_norm_singularity_correction * eps_out * E_fwd_norm * E_adj_norm
        vjps = contrib_E

        # compute the tangential H contribution to the gradient (the normal H contribution
        # is 0 for PEC)
        H_fwd_perp1 = self._project_in_basis(H_fwd_at_coords, basis_vector=perps1)
        H_adj_perp1 = self._project_in_basis(H_adj_at_coords, basis_vector=perps1)

        H_fwd_perp2 = self._project_in_basis(H_fwd_at_coords, basis_vector=perps2)
        H_adj_perp2 = self._project_in_basis(H_adj_at_coords, basis_vector=perps2)

        H_der_perp1 = H_perp_singularity_correction * H_fwd_perp1 * H_adj_perp1
        H_der_perp2 = H_perp_singularity_correction * H_fwd_perp2 * H_adj_perp2

        H_integration_components = (H_der_perp1, H_der_perp2)
        if pec_line_integration:
            # if we are integrating along the line, we choose the H component normal to
            # the edge which corresponds to a surface current along the edge whereas the other
            # tangential component corresponds to a surface current along the flat dimension.
            H_integration_components = tuple(
                H_comp for idx, H_comp in enumerate(H_integration_components) if flat_perp_dims[idx]
            )

        # for each of the tangential components we are integrating the H fields over,
        # adjust weighting to account for pre-weighting of the source by `EPSILON_0`
        # and multiply by appropriate `MU_0` factor
        for H_perp in H_integration_components:
            contrib_H = MU_0 * H_perp / EPSILON_0
            vjps += contrib_H

        return vjps

    @staticmethod
    def _prepare_epsilon(eps: EpsType) -> np.ndarray:
        """Prepare epsilon values for multi-frequency.

        For FreqDataArray, extracts values and broadcasts to shape (1, n_freqs).
        For scalar values, broadcasts to shape (1, 1) for consistency with multi-frequency.
        """
        if isinstance(eps, FreqDataArray):
            # data is already sliced, just extract values
            eps_values = eps.values
            # shape: (n_freqs,) - need to broadcast to (1, n_freqs)
            return eps_values[np.newaxis, :]
        else:
            # scalar value - broadcast to (1, 1)
            return np.array([[eps]])

    @staticmethod
    def _project_in_basis(
        field_components: dict[str, np.ndarray],
        basis_vector: np.ndarray,
    ) -> np.ndarray:
        """Project 3D field components onto a basis vector.

        Parameters
        ----------
        field_components : dict[str, np.ndarray]
            Dictionary with keys like "Ex", "Ey", "Ez" or "Dx", "Dy", "Dz" containing field values.
            Values have shape (N, F) where F is the number of frequencies.
        basis_vector : np.ndarray
            (N, 3) array of basis vectors, one per evaluation point.

        Returns
        -------
        np.ndarray
            Projected field values with shape (N, F).
        """
        prefix = next(iter(field_components.keys()))[0]
        field_matrix = np.stack([field_components[f"{prefix}{dim}"] for dim in "xyz"], axis=0)

        # always expect (3, N, F) shape, transpose to (N, 3, F)
        field_matrix = np.transpose(field_matrix, (1, 0, 2))
        return np.einsum("ij...,ij->i...", field_matrix, basis_vector)

    def adaptive_vjp_spacing(
        self,
        wl_fraction: float = DEFAULT_WAVELENGTH_FRACTION,
        min_allowed_spacing_fraction: float = MINIMUM_SPACING_FRACTION,
    ) -> float:
        """Compute adaptive spacing for finite-difference gradient evaluation.

        Determines an appropriate spatial resolution based on the material
        properties and electromagnetic wavelength/skin depth.

        Parameters
        ----------
        wl_fraction : float = 0.1
            Fraction of wavelength/skin depth to use as spacing.
        min_allowed_spacing_fraction : float = 1e-2
            Minimum allowed spacing fraction of free space wavelength to
            prevent numerical issues.

        Returns
        -------
        float
            Adaptive spacing value for gradient evaluation.
        """
        # handle FreqDataArray or scalar eps_in
        if isinstance(self.eps_in, FreqDataArray):
            eps_real = np.asarray(self.eps_in.values, dtype=np.complex128).real
        else:
            eps_real = np.asarray(self.eps_in, dtype=np.complex128).real

        dx_candidates = []
        max_frequency = np.max(self.frequencies)

        # wavelength-based sampling for dielectrics
        if np.any(eps_real > 0):
            eps_max = eps_real[eps_real > 0].max()
            lambda_min = self.wavelength_min / np.sqrt(eps_max)
            dx_candidates.append(wl_fraction * lambda_min)

        # skin depth sampling for metals
        if np.any(eps_real <= 0):
            omega = 2 * np.pi * max_frequency
            eps_neg = eps_real[eps_real <= 0]
            delta_min = C_0 / (omega * np.sqrt(np.abs(eps_neg).max()))
            dx_candidates.append(wl_fraction * delta_min)

        computed_spacing = min(dx_candidates)
        min_allowed_spacing = self.wavelength_min * min_allowed_spacing_fraction

        if computed_spacing < min_allowed_spacing:
            log.warning(
                f"Based on the material, the adaptive spacing for integrating the polyslab surface "
                f"would be {computed_spacing:.3e} μm. The spacing has been clipped to {min_allowed_spacing:.3e} μm "
                f"to prevent a performance degradation.",
                log_once=True,
            )
        return max(computed_spacing, min_allowed_spacing)

    @property
    def wavelength_min(self) -> float:
        return C_0 / np.max(self.frequencies)

    @property
    def wavelength_max(self) -> float:
        return C_0 / np.min(self.frequencies)


def integrate_within_bounds(arr: xr.DataArray, dims: list[str], bounds: Bound) -> xr.DataArray:
    """Integrate a data array within specified spatial bounds.

    Clips the integration domain to the specified bounds and performs
    numerical integration using the trapezoidal rule.

    Parameters
    ----------
    arr : xr.DataArray
        Data array to integrate.
    dims : list[str]
        Dimensions to integrate over (e.g., ['x', 'y', 'z']).
    bounds : Bound
        Integration bounds as [[xmin, ymin, zmin], [xmax, ymax, zmax]].

    Returns
    -------
    xr.DataArray
        Result of integration with specified dimensions removed.

    Notes
    -----
    - Coordinates outside bounds are clipped, effectively setting dL=0
    - Only integrates dimensions with more than one coordinate point
    - Uses xarray's integrate method (trapezoidal rule)
    """
    bounds = np.asarray(bounds).T
    all_coords = {}

    for dim, (bmin, bmax) in zip(dims, bounds):
        bmin = get_static(bmin)
        bmax = get_static(bmax)

        # clip coordinates to bounds (sets dL=0 outside bounds)
        coord_values = arr.coords[dim].data
        all_coords[dim] = np.clip(coord_values, bmin, bmax)

    _arr = arr.assign_coords(**all_coords)

    # only integrate dimensions with multiple points
    dims_integrate = [dim for dim in dims if len(_arr.coords[dim]) > 1]
    return _arr.integrate(coord=dims_integrate)


__all__ = [
    "DerivativeInfo",
    "integrate_within_bounds",
]



"""Lumped port specialization with a rectangular geometry."""

from __future__ import annotations

from typing import Optional

import numpy as np
import pydantic.v1 as pd

from tidy3d.components.base import cached_property
from tidy3d.components.data.data_array import FreqDataArray
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.geometry.base import Box
from tidy3d.components.geometry.utils import (
    SnapBehavior,
    SnapLocation,
    SnappingSpec,
    snap_box_to_grid,
)
from tidy3d.components.geometry.utils_2d import increment_float
from tidy3d.components.grid.grid import Grid, YeeGrid
from tidy3d.components.lumped_element import LinearLumpedElement, LumpedResistor, RLCNetwork
from tidy3d.components.monitor import FieldMonitor
from tidy3d.components.source.current import UniformCurrentSource
from tidy3d.components.source.time import GaussianPulse
from tidy3d.components.types import Axis, FreqArray, LumpDistType
from tidy3d.components.validators import assert_line_or_plane
from tidy3d.exceptions import SetupError, ValidationError
from tidy3d.plugins.microwave import CurrentIntegralAxisAligned, VoltageIntegralAxisAligned

from .base_lumped import AbstractLumpedPort


class LumpedPort(AbstractLumpedPort, Box):
    """Class representing a single rectangular lumped port.

    Example
    -------
    >>> port1 = LumpedPort(center=(0, 0, 0),
    ...             size=(0, 1, 2),
    ...             voltage_axis=2,
    ...             name="port_1",
    ...             impedance=50
    ...         ) # doctest: +SKIP

    See Also
    --------
    :class:`.LinearLumpedElement`
        The lumped element representing the load of the port.
    """

    voltage_axis: Axis = pd.Field(
        ...,
        title="Voltage Integration Axis",
        description="Specifies the axis along which the E-field line integral is performed when "
        "computing the port voltage. The integration axis must lie in the plane of the port.",
    )

    snap_perimeter_to_grid: bool = pd.Field(
        True,
        title="Snap Perimeter to Grid",
        description="When enabled, the perimeter of the port is snapped to the simulation grid, "
        "which improves accuracy when the number of grid cells is low within the element. A :class:`LumpedPort` "
        "is always snapped to the grid along its injection axis.",
    )

    dist_type: LumpDistType = pd.Field(
        "on",
        title="Distribute Type",
        description="Optional field that is passed directly to the :class:`.LinearLumpedElement` used to model the port's load. "
        "When set to ``on``, the network portion of the lumped port, including the source, is distributed"
        "across the entirety of the lumped element's bounding box. When set to ``off``, the network "
        "portion of the lumped port is restricted to one cell and PEC connections are used to "
        "connect the network cell to the edges of the lumped element. A third option exists "
        "``laterally_only``, where the network portion is only distributed along the lateral axis of "
        "the lumped port.",
    )

    _line_plane_validator = assert_line_or_plane()

    @cached_property
    def injection_axis(self):
        """Injection axis of the port."""
        return self.size.index(0.0)

    @pd.validator("voltage_axis", always=True)
    def _voltage_axis_in_plane(cls, val, values):
        """Ensure voltage integration axis is in the port's plane."""
        size = values.get("size")
        if val == size.index(0.0):
            raise ValidationError("'voltage_axis' must lie in the port's plane.")
        return val

    @cached_property
    def current_axis(self) -> Axis:
        """Integration axis for computing the port current via the magnetic field."""
        return 3 - self.injection_axis - self.voltage_axis

    def to_source(
        self, source_time: GaussianPulse, snap_center: Optional[float] = None, grid: Grid = None
    ) -> UniformCurrentSource:
        """Create a current source from the lumped port."""
        if grid:
            # This will included any snapping behavior the load undergoes
            load_box = self._to_load_box(grid=grid)
            center = load_box.center
            size = load_box.size
        else:
            # Discretized source amps are manually zeroed out later if they
            # fall on Yee grid locations outside the analytical source region.
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            size = self.size

        component = "xyz"[self.voltage_axis]
        return UniformCurrentSource(
            center=center,
            size=size,
            source_time=source_time,
            polarization=f"E{component}",
            name=self.name,
            interpolate=True,
            confine_to_bounds=True,
        )

    def to_load(self, snap_center: Optional[float] = None) -> LumpedResistor:
        """Create a load resistor from the lumped port."""
        # 2D materials are currently snapped to the grid, so snapping here is not needed.
        # It is done here so plots of the simulation will more accurately portray the setup
        center = list(self.center)
        if snap_center:
            center[self.injection_axis] = snap_center

        network = RLCNetwork(resistance=np.real(self.impedance))
        return LinearLumpedElement(
            center=center,
            size=self.size,
            num_grid_cells=self.num_grid_cells,
            network=network,
            name=f"{self.name}_resistor",
            voltage_axis=self.voltage_axis,
            snap_perimeter_to_grid=self.snap_perimeter_to_grid,
            dist_type=self.dist_type,
            enable_snapping_points=self.enable_snapping_points,
        )

    def to_voltage_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> FieldMonitor:
        """Field monitor to compute port voltage."""
        if grid:
            voltage_box = self._to_voltage_box(grid)
            center = voltage_box.center
            size = voltage_box.size
        else:
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            # Size of voltage monitor can essentially be 1D from ground to signal conductor
            size = list(self.size)
            size[self.injection_axis] = 0.0
            size[self.current_axis] = 0.0

        e_component = "xyz"[self.voltage_axis]
        # Create a voltage monitor
        return FieldMonitor(
            center=center,
            size=size,
            freqs=freqs,
            fields=[f"E{e_component}"],
            name=self._voltage_monitor_name,
            colocate=False,
        )

    def to_current_monitor(
        self, freqs: FreqArray, snap_center: Optional[float] = None, grid: Grid = None
    ) -> FieldMonitor:
        """Field monitor to compute port current."""
        if grid:
            current_box = self._to_current_box(grid)
            center = current_box.center
            size = current_box.size
        else:
            center = list(self.center)
            if snap_center:
                center[self.injection_axis] = snap_center
            # Size of current monitor needs to encompass the current carrying 2D sheet
            # Needs to have a nonzero thickness so a closed loop of gridpoints around
            # the 2D sheet can be formed
            dl = 2 * (
                increment_float(center[self.injection_axis], 1.0) - center[self.injection_axis]
            )
            size = list(self.size)
            size[self.injection_axis] = dl
            size[self.voltage_axis] = 0.0

        h_component = "xyz"[self.current_axis]
        h_cap_component = "xyz"[self.injection_axis]
        # Create a current monitor
        return FieldMonitor(
            center=center,
            size=size,
            freqs=freqs,
            fields=[f"H{h_component}", f"H{h_cap_component}"],
            name=self._current_monitor_name,
            colocate=False,
        )

    def compute_voltage(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute voltage across the port."""
        voltage_box = self._to_voltage_box(sim_data.simulation.grid)
        field_data = sim_data[self._voltage_monitor_name]
        voltage_integral = VoltageIntegralAxisAligned(
            center=voltage_box.center,
            size=voltage_box.size,
            extrapolate_to_endpoints=True,
            snap_path_to_grid=True,
            sign="+",
        )
        voltage = voltage_integral.compute_voltage(field_data)
        # Return data array of voltage with coordinates of frequency
        return voltage

    def compute_current(self, sim_data: SimulationData) -> FreqDataArray:
        """Helper to compute current flowing through the port."""
        # Diagram of contour integral, dashed line indicates location of sheet resistance
        # and electric field used for voltage computation. Voltage axis is out-of-page.
        #
        #                                    current_axis = ->
        #                                    injection_axis = ^
        #
        #                  |                   h2_field ->             |
        #    h_cap_minus ^  -------------------------------------------  h_cap_plus ^
        #                  |                   h1_field ->             |

        field_data = sim_data[self._current_monitor_name]
        current_box = self._to_current_box(sim_data.simulation.grid)

        # H field is continuous at integral bounds, so extrapolation is turned off
        I_integral = CurrentIntegralAxisAligned(
            center=current_box.center,
            size=current_box.size,
            sign="+",
            extrapolate_to_endpoints=True,
            snap_contour_to_grid=True,
        )
        return I_integral.compute_current(field_data)

    def _check_grid_size(self, yee_grid: YeeGrid):
        """Raises :class:`SetupError` if the grid is too coarse at port locations"""
        e_component = "xyz"[self.voltage_axis]
        e_yee_grid = yee_grid.grid_dict[f"E{e_component}"]
        coords = e_yee_grid.to_dict[e_component]
        min_bound = self.bounds[0][self.voltage_axis]
        max_bound = self.bounds[1][self.voltage_axis]
        coords_within_port = np.any(np.logical_and(coords > min_bound, coords < max_bound))
        if not coords_within_port:
            raise SetupError(
                f"Grid is too coarse along '{e_component}' direction for the lumped port "
                f"at location '{self.center}'. Either set the port's 'num_grid_cells' to "
                f"a nonzero integer or modify the 'GridSpec'."
            )

    def _to_load_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the exact location of the load,
        after it is snapped to the grid."""
        load = self.to_load()
        # This will included any snapping behavior the load undergoes
        load_box = load._create_box_for_network(grid=grid)
        return load_box

    def _to_voltage_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the location of the
        path integral for computing voltage."""
        load_box = self._to_load_box(grid=grid)
        size = list(load_box.size)
        size[self.current_axis] = 0
        size[self.injection_axis] = 0
        voltage_box = Box(center=load_box.center, size=size)
        return voltage_box

    def _to_current_box(self, grid: Grid) -> Box:
        """Helper to get a ``Box`` representing the location of the
        path integral for computing current."""
        load_box = self._to_load_box(grid=grid)
        size = list(load_box.size)
        size[self.voltage_axis] = 0
        current_box = Box(center=load_box.center, size=size)
        # Snap the current contour integral to the nearest magnetic field positions
        # that enclose the load box/sheet resistance
        snap_location = [SnapLocation.Center] * 3
        snap_behavior = [SnapBehavior.Expand] * 3
        snap_behavior[self.voltage_axis] = SnapBehavior.Off
        snap_spec = SnappingSpec(location=snap_location, behavior=snap_behavior)
        current_box = snap_box_to_grid(grid, current_box, snap_spec)
        return current_box



from __future__ import annotations

from typing import Union

from tidy3d.components.data.monitor_data import ModeSolverData
from tidy3d.components.data.sim_data import SimulationData
from tidy3d.components.eme.data.sim_data import EMESimulationData
from tidy3d.components.eme.simulation import EMESimulation
from tidy3d.components.mode.data.sim_data import ModeSimulationData
from tidy3d.components.mode.simulation import ModeSimulation
from tidy3d.components.simulation import Simulation
from tidy3d.components.tcad.data.sim_data import (
    HeatChargeSimulationData,
    HeatSimulationData,
    VolumeMesherData,
)
from tidy3d.components.tcad.mesher import VolumeMesher
from tidy3d.components.tcad.simulation.heat import HeatSimulation
from tidy3d.components.tcad.simulation.heat_charge import HeatChargeSimulation
from tidy3d.plugins.mode.mode_solver import ModeSolver

SimulationType = Union[
    Simulation,
    HeatChargeSimulation,
    HeatSimulation,
    EMESimulation,
    ModeSolver,
    ModeSimulation,
    VolumeMesher,
]
SimulationDataType = Union[
    SimulationData,
    HeatChargeSimulationData,
    HeatSimulationData,
    EMESimulationData,
    ModeSolverData,
    ModeSimulationData,
    VolumeMesherData,
]



# container for everything defining the inverse design

from __future__ import annotations

import abc
import typing

import autograd.numpy as anp
import numpy as np
import pydantic.v1 as pd

import tidy3d as td
from tidy3d.components.autograd import get_static
from tidy3d.exceptions import ValidationError
from tidy3d.plugins.expressions.metrics import Metric, generate_validation_data
from tidy3d.plugins.expressions.types import ExpressionType

from .base import InvdesBaseModel
from .region import DesignRegionType
from .validators import check_pixel_size

PostProcessFnType = typing.Callable[[td.SimulationData], float]


class AbstractInverseDesign(InvdesBaseModel, abc.ABC):
    """Container for an inverse design problem."""

    design_region: DesignRegionType = pd.Field(
        ...,
        title="Design Region",
        description="Region within which we will optimize the simulation.",
    )

    task_name: str = pd.Field(
        ...,
        title="Task Name",
        description="Task name to use in the objective function when running the ``JaxSimulation``.",
    )

    verbose: bool = pd.Field(
        False,
        title="Task Verbosity",
        description="If ``True``, will print the regular output from ``web`` functions.",
    )

    metric: typing.Optional[ExpressionType] = pd.Field(
        None,
        title="Objective Metric",
        description="Serializable expression defining the objective function.",
    )

    def make_objective_fn(
        self, post_process_fn: typing.Optional[typing.Callable] = None, maximize: bool = True
    ) -> typing.Callable[[anp.ndarray], tuple[float, dict]]:
        """Construct the objective function for this InverseDesign object."""

        if (post_process_fn is None) and (self.metric is None):
            raise ValueError("Either 'post_process_fn' or 'metric' must be provided.")

        if (post_process_fn is not None) and (self.metric is not None):
            raise ValueError("Provide only one of 'post_process_fn' or 'metric', not both.")

        direction_multiplier = 1 if maximize else -1

        def objective_fn(params: anp.ndarray, aux_data: typing.Optional[dict] = None) -> float:
            """Full objective function."""
            data = self.to_simulation_data(params=params)

            if self.metric is None:
                post_process_val = post_process_fn(data)
            elif isinstance(data, td.SimulationData):
                post_process_val = self.metric.evaluate(data)
            elif getattr(data, "type", None) == "BatchData":
                raise NotImplementedError("Metrics currently do not support 'BatchData'")
            else:
                raise ValueError(f"Invalid data type: {type(data)}")

            penalty_value = self.design_region.penalty_value(params)
            objective_fn_val = direction_multiplier * post_process_val - penalty_value

            # Store auxiliary data if provided
            if aux_data is not None:
                aux_data["penalty"] = get_static(penalty_value)
                aux_data["post_process_val"] = get_static(post_process_val)
                aux_data["objective_fn_val"] = get_static(objective_fn_val) * direction_multiplier
                if isinstance(data, td.SimulationData):
                    aux_data["sim_data"] = data.to_static()
                else:
                    aux_data["sim_data"] = {k: v.to_static() for k, v in data.items()}
                aux_data["params"] = params

            return objective_fn_val

        return objective_fn

    @property
    def initial_simulation(self) -> td.Simulation:
        """Return a simulation with the initial design region parameters."""
        initial_params = self.design_region.initial_parameters
        return self.to_simulation(initial_params)

    def run(self, simulation, **kwargs) -> td.SimulationData:
        """Run a single tidy3d simulation."""
        from tidy3d.web import run

        kwargs.setdefault("verbose", self.verbose)
        kwargs.setdefault("task_name", self.task_name)
        return run(simulation, **kwargs)

    def run_async(self, simulations, **kwargs) -> web.BatchData:  # noqa: F821
        """Run a batch of tidy3d simulations."""
        from tidy3d.web import run_async

        kwargs.setdefault("verbose", self.verbose)
        return run_async(simulations, **kwargs)


class InverseDesign(AbstractInverseDesign):
    """Container for an inverse design problem."""

    simulation: td.Simulation = pd.Field(
        ...,
        title="Base Simulation",
        description="Simulation without the design regions or monitors used in the objective fn.",
    )

    output_monitor_names: tuple[str, ...] = pd.Field(
        None,
        title="Output Monitor Names",
        description="Optional names of monitors whose data the differentiable output depends on."
        "If this field is left ``None``, the plugin will try to add all compatible monitors to "
        "``JaxSimulation.output_monitors``. While this will work, there may be warnings if the "
        "monitors are not compatible with the ``adjoint`` plugin, for example if there are "
        "``FieldMonitor`` instances with ``.colocate != False``.",
    )

    _check_sim_pixel_size = check_pixel_size("simulation")

    @pd.root_validator(pre=False)
    def _validate_model(cls, values: dict) -> dict:
        cls._validate_metric(values)
        return values

    @staticmethod
    def _validate_metric(values: dict) -> dict:
        metric_expr = values.get("metric")
        if not metric_expr:
            return values
        simulation = values.get("simulation")
        for metric in metric_expr.filter(Metric):
            InverseDesign._validate_metric_monitor_name(metric, simulation)
            InverseDesign._validate_metric_mode_index(metric, simulation)
            InverseDesign._validate_metric_f(metric, simulation)
        InverseDesign._validate_metric_data(metric_expr, simulation)
        return values

    @staticmethod
    def _validate_metric_monitor_name(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the monitor name of the metric exists in the simulation."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if monitor is None:
            raise ValidationError(
                f"Monitor named '{metric.monitor_name}' associated with the metric not found in the simulation monitors."
            )

    @staticmethod
    def _validate_metric_mode_index(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the mode index of the metric is within the bounds of the monitor's ``ModeSpec.num_modes``."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if metric.mode_index >= monitor.mode_spec.num_modes:
            raise ValidationError(
                f"Mode index '{metric.mode_index}' for metric associated with monitor "
                f"'{metric.monitor_name}' is out of bounds. "
                f"Maximum allowed mode index is '{monitor.mode_spec.num_modes - 1}'."
            )

    @staticmethod
    def _validate_metric_f(metric: Metric, simulation: td.Simulation) -> None:
        """Validate that the frequencies of the metric are present in the monitor."""
        monitor = next((m for m in simulation.monitors if m.name == metric.monitor_name), None)
        if metric.f is not None:
            metric_f_list = [metric.f] if isinstance(metric.f, float) else metric.f
            if len(metric_f_list) != 1:
                raise ValidationError("Only a single frequency is supported for the metric.")
            for freq in metric_f_list:
                if not any(np.isclose(freq, monitor.freqs, atol=1.0)):
                    raise ValidationError(
                        f"Frequency '{freq}' for metric associated with monitor "
                        f"'{metric.monitor_name}' not found in monitor frequencies."
                    )
        else:
            if len(monitor.freqs) != 1:
                raise ValidationError(
                    f"Monitor '{metric.monitor_name}' must contain only a single frequency when metric.f is None."
                )

    @staticmethod
    def _validate_metric_data(expr: ExpressionType, simulation: td.Simulation) -> None:
        """Validate that expression can be evaluated and returns a real scalar."""
        data = generate_validation_data(expr)
        try:
            result = expr(data)
        except Exception as e:
            raise ValidationError(f"Failed to evaluate the metric expression: {e!s}") from e
        if len(np.ravel(result)) > 1:
            raise ValidationError(
                f"The expression must return a scalar value or an array of length 1 (got {result})."
            )
        if not np.all(np.isreal(result)):
            raise ValidationError(
                f"The expression must return a real (not complex) value (got {result})."
            )

    def is_output_monitor(self, monitor: td.Monitor) -> bool:
        """Whether a monitor is added to the ``JaxSimulation`` as an ``output_monitor``."""

        output_mnt_types = td.components.simulation.OutputMonitorTypes

        if self.output_monitor_names is None:
            return any(isinstance(monitor, mnt_type) for mnt_type in output_mnt_types)

        return monitor.name in self.output_monitor_names

    def separate_output_monitors(self, monitors: tuple[td.Monitor]) -> dict:
        """Separate monitors into output_monitors and regular monitors."""

        monitor_fields = {"monitors": [], "output_monitors": []}

        for monitor in monitors:
            key = "output_monitors" if self.is_output_monitor(monitor) else "monitors"
            monitor_fields[key].append(monitor)

        return monitor_fields

    def to_simulation(self, params: anp.ndarray) -> td.Simulation:
        """Convert the ``InverseDesign`` to a corresponding ``td.Simulation`` with traced fields."""

        # construct the design region to a regular structure
        design_region_structure = self.design_region.to_structure(params)

        # construct mesh override structures and a new grid spec, if applicable
        grid_spec = self.simulation.grid_spec
        mesh_override_structure = self.design_region.mesh_override_structure
        if mesh_override_structure:
            override_structures = list(self.simulation.grid_spec.override_structures)
            override_structures += [mesh_override_structure]
            grid_spec = grid_spec.updated_copy(override_structures=override_structures)

        return self.simulation.updated_copy(
            structures=[*list(self.simulation.structures), design_region_structure],
            grid_spec=grid_spec,
        )

    def to_simulation_data(self, params: anp.ndarray, **kwargs) -> td.SimulationData:
        """Convert the ``InverseDesign`` to a ``td.Simulation`` and run it."""
        simulation = self.to_simulation(params=params)
        return self.run(simulation, **kwargs)


class InverseDesignMulti(AbstractInverseDesign):
    """``InverseDesign`` with multiple simulations and corresponding postprocess functions."""

    simulations: tuple[td.Simulation, ...] = pd.Field(
        ...,
        title="Base Simulations",
        description="Set of simulation without the design regions or monitors used in the objective fn.",
    )

    output_monitor_names: tuple[typing.Union[tuple[str, ...], None], ...] = pd.Field(
        None,
        title="Output Monitor Names",
        description="Optional names of monitors whose data the differentiable output depends on."
        "If this field is left ``None``, the plugin will try to add all compatible monitors to "
        "``JaxSimulation.output_monitors``. While this will work, there may be warnings if the "
        "monitors are not compatible with the ``adjoint`` plugin, for example if there are "
        "``FieldMonitor`` instances with ``.colocate != False``.",
    )

    _check_sim_pixel_size = check_pixel_size("simulations")

    @pd.root_validator()
    def _check_lengths(cls, values):
        """Check the lengths of all of the multi fields."""

        keys = ("simulations", "post_process_fns", "output_monitor_names", "override_structure_dl")
        multi_dict = {key: values.get(key) for key in keys}
        sizes = {key: len(val) for key, val in multi_dict.items() if val is not None}

        if len(set(sizes.values())) != 1:
            raise ValueError(
                f"'MultiInverseDesign' requires that the fields {keys} must either "
                "have the same length or be left ``None``, if optional. Given fields with "
                "corresponding sizes of '{sizes}'."
            )

        return values

    @property
    def task_names(self) -> list[str]:
        """Task names associated with each of the simulations."""
        return [f"{self.task_name}_{i}" for i in range(len(self.simulations))]

    @property
    def designs(self) -> list[InverseDesign]:
        """List of individual ``InverseDesign`` objects corresponding to this instance."""

        designs_list = []
        for i, (task_name, sim) in enumerate(zip(self.task_names, self.simulations)):
            des_i = InverseDesign(
                design_region=self.design_region,
                simulation=sim,
                verbose=self.verbose,
                task_name=task_name,
            )
            if self.output_monitor_names is not None:
                des_i = des_i.updated_copy(output_monitor_names=self.output_monitor_names[i])

            designs_list.append(des_i)

        return designs_list

    def to_simulation(self, params: anp.ndarray) -> dict[str, td.Simulation]:
        """Convert the ``InverseDesign`` to a corresponding dict of ``td.Simulation``s."""
        simulation_list = [design.to_simulation(params) for design in self.designs]
        return dict(zip(self.task_names, simulation_list))

    def to_simulation_data(self, params: anp.ndarray, **kwargs) -> web.BatchData:  # noqa: F821
        """Convert the ``InverseDesignMulti`` to a set of ``td.Simulation``s and run async."""
        simulations = self.to_simulation(params)
        return self.run_async(simulations, **kwargs)


InverseDesignType = typing.Union[InverseDesign, InverseDesignMulti]



"""Custom Tidy3D exceptions"""

from __future__ import annotations

from typing import Optional

from .core_config import get_logger


class WebError(Exception):
    """Any error in tidy3d"""

    def __init__(self, message: Optional[str] = None):
        """Log just the error message and then raise the Exception."""
        log = get_logger()
        super().__init__(message)
        log.error(message)


class WebNotFoundError(WebError):
    """A generic error indicating an HTTP 404 (resource not found)."""



from __future__ import annotations

import itertools

import autograd.numpy as anp
import numpy as np
from autograd.extend import defjvp, defvjp, primitive
from autograd.numpy.numpy_jvps import broadcast
from autograd.numpy.numpy_vjps import unbroadcast_f
from numpy.typing import NDArray

from .types import InterpolationType


def _evaluate_nearest(
    indices: NDArray[np.int64], norm_distances: NDArray[np.float64], values: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Perform nearest neighbor interpolation in an n-dimensional space.

    This function determines the nearest neighbor in a grid for a given point
    and returns the corresponding value from the input array.

    Parameters
    ----------
    indices : np.ndarray[np.int64]
        Indices of the lower bounds of the grid cell containing the interpolation point.
    norm_distances : np.ndarray[np.float64]
        Normalized distances from the lower bounds of the grid cell to the
        interpolation point, for each dimension.
    values : np.ndarray[np.float64]
        The n-dimensional array of values to interpolate from.

    Returns
    -------
    np.ndarray[np.float64]
        The value of the nearest neighbor to the interpolation point.
    """
    idx_res = tuple(anp.where(yi <= 0.5, i, i + 1) for i, yi in zip(indices, norm_distances))
    return values[idx_res]


def _evaluate_linear(
    indices: NDArray[np.int64], norm_distances: NDArray[np.float64], values: NDArray[np.float64]
) -> NDArray[np.float64]:
    """Perform linear interpolation in an n-dimensional space.

    This function calculates the linearly interpolated value at a point in an
    n-dimensional grid, given the indices of the surrounding grid points and
    the normalized distances to these points.
    The multi-linear interpolation is implemented by computing a weighted
    average of the values at the vertices of the hypercube surrounding the
    interpolation point.

    Parameters
    ----------
    indices : np.ndarray[np.int64]
        Indices of the lower bounds of the grid cell containing the interpolation point.
    norm_distances : np.ndarray[np.float64]
        Normalized distances from the lower bounds of the grid cell to the
        interpolation point, for each dimension.
    values : np.ndarray[np.float64]
        The n-dimensional array of values to interpolate from.

    Returns
    -------
    np.ndarray[np.float64]
        The interpolated value at the desired point.
    """
    # Create a slice object for broadcasting over trailing dimensions
    _slice = (slice(None),) + (None,) * (values.ndim - len(indices))

    # Prepare iterables for lower and upper bounds of the hypercube
    ix = zip(indices, (1 - yi for yi in norm_distances))
    iy = zip((i + 1 for i in indices), norm_distances)

    # Initialize the result
    value = anp.zeros(1)

    # Iterate over all vertices of the hypercube
    for h in itertools.product(*zip(ix, iy)):
        edge_indices, weights = zip(*h)

        # Compute the weight for this vertex
        weight = anp.ones(1)
        for w in weights:
            weight = weight * w

        # Compute the contribution of this vertex and add it to the result
        term = values[edge_indices] * weight[_slice]
        value = value + term

    return value


def interpn(
    points: tuple[NDArray[np.float64], ...],
    values: NDArray[np.float64],
    xi: tuple[NDArray[np.float64], ...],
    *,
    method: InterpolationType = "linear",
    **kwargs,
) -> NDArray[np.float64]:
    """Interpolate over a rectilinear grid in arbitrary dimensions.

    This function mirrors the interface of `scipy.interpolate.interpn` but is differentiable with autograd.

    Parameters
    ----------
    points : tuple[np.ndarray[np.float64], ...]
        The points defining the rectilinear grid in n dimensions.
    values : np.ndarray[np.float64]
        The data values on the rectilinear grid.
    xi : tuple[np.ndarray[np.float64], ...]
        The coordinates to sample the gridded data at.
    method : InterpolationType = "linear"
        The method of interpolation to perform. Supported are "linear" and "nearest".

    Returns
    -------
    np.ndarray[np.float64]
        The interpolated values.

    Raises
    ------
    ValueError
        If the interpolation method is not supported.

    See Also
    --------
    `scipy.interpolate.interpn <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interpn.html>`_
    """
    from scipy.interpolate import RegularGridInterpolator

    if method == "nearest":
        interp_fn = _evaluate_nearest
    elif method == "linear":
        interp_fn = _evaluate_linear
    else:
        raise ValueError(f"Unsupported interpolation method: {method}")

    if kwargs.get("fill_value") == "extrapolate":
        itrp = RegularGridInterpolator(
            points, values, method=method, fill_value=None, bounds_error=False
        )
    else:
        itrp = RegularGridInterpolator(points, values, method=method)

    # Prepare the grid for interpolation
    # This step reshapes the grid, checks for NaNs and out-of-bounds values
    # It returns:
    #   - reshaped grid
    #   - original shape
    #   - number of dimensions
    #   - boolean array indicating NaN positions
    #   - (discarded) boolean array for out-of-bounds values
    xi, shape, ndim, nans, _ = itrp._prepare_xi(xi)

    # Find the indices of the grid cells containing the interpolation points
    # and calculate the normalized distances (ranging from 0 at lower grid point to 1
    # at upper grid point) within these cells
    indices, norm_distances = itrp._find_indices(xi.T)

    result = interp_fn(indices, norm_distances, values)
    nans = anp.reshape(nans, (-1,) + (1,) * (result.ndim - 1))
    result = anp.where(nans, np.nan, result)
    return anp.reshape(result, shape[:-1] + values.shape[ndim:])


def trapz(y: NDArray, x: NDArray = None, dx: float = 1.0, axis: int = -1) -> float:
    """
    Integrate along the given axis using the composite trapezoidal rule.

    Parameters
    ----------
    y : np.ndarray
        Input array to integrate.
    x : np.ndarray = None
        The sample points corresponding to the y values. If None, the sample points are assumed to be evenly spaced
        with spacing `dx`.
    dx : float = 1.0
        The spacing between sample points when `x` is None. Default is 1.0.
    axis : int = -1
        The axis along which to integrate. Default is the last axis.

    Returns
    -------
    float
        Definite integral as approximated by the trapezoidal rule.
    """
    if x is None:
        d = dx
    elif x.ndim == 1:
        d = np.diff(x)
        shape = [1] * y.ndim
        shape[axis] = d.shape[0]
        d = np.reshape(d, shape)
    else:
        d = np.diff(x, axis=axis)

    slice1 = [slice(None)] * y.ndim
    slice2 = [slice(None)] * y.ndim
    slice1[axis] = slice(1, None)
    slice2[axis] = slice(None, -1)

    return anp.sum((y[tuple(slice1)] + y[tuple(slice2)]) * d / 2, axis=axis)


@primitive
def _add_at(x: NDArray, indices_x: tuple, y: NDArray) -> NDArray:
    """
    Add values to specified indices of an array.

    Autograd requires that arguments to primitives are passed in positionally.
    ``add_at`` is the public-facing wrapper for this function,
    which allows keyword arguments in case users pass in kwargs.
    """
    out = np.copy(x)  # Copy to preserve 'x' for gradient computation
    out[tuple(indices_x)] += y
    return out


defvjp(
    _add_at,
    lambda ans, x, indices_x, y: unbroadcast_f(x, lambda g: g),
    lambda ans, x, indices_x, y: lambda g: g[tuple(indices_x)],
    argnums=(0, 2),
)

defjvp(
    _add_at,
    lambda g, ans, x, indices_x, y: broadcast(g, ans),
    lambda g, ans, x, indices_x, y: _add_at(anp.zeros_like(ans), indices_x, g),
    argnums=(0, 2),
)


def add_at(x: NDArray, indices_x: tuple, y: NDArray) -> NDArray:
    """
    Add values to specified indices of an array.

    This function creates a copy of the input array `x`, adds the values from `y` to the specified
    indices `indices_x`, and returns the modified array.

    Parameters
    ----------
    x : np.ndarray
        Input array to which values will be added.
    indices_x : tuple
        Indices of `x` where values from `y` will be added.
    y : np.ndarray
        Values to add to the specified indices of `x`.

    Returns
    -------
    np.ndarray
        The modified array with values added at the specified indices.
    """
    return _add_at(x, indices_x, y)


__all__ = [
    "add_at",
    "interpn",
    "trapz",
]
